"""Datastruc module.

This module contains definitions of the datastructure the
SMLM dataitem will be parsed as.

"""

import numpy as np
import napari
import matplotlib.pyplot as plt
import polars as pl
import pyarrow.parquet as pq
import ast
import os
import json

_interpolate = {
    "log2": lambda d: np.log2(d),
    "log10": lambda d: np.log10(d),
    "linear": lambda d: d,
}


class item:
    """smlm datastructure.
    
    This is the basic datastructure which will contain all the information
    from a point set that is needed

    Attributes:
        name (string) : Contains the name of the item
        df (polars dataframe): Dataframe with the data contained, containing
            columns: 'channel' 'frame' 'x' 'y' 'z'.
            If manual annotation is done an additional column 'gt_label'
            will be present
        dim (int): Dimensions of data
        channels (list): list of ints, representing channels user wants
            to consider in the original data
        histo (dict): Dictionary of 2D or 3D arrays. Each key corresponds
            to the channel for which the histogram
            contains the relevant binned data, in form [X,Y,Z]
            i.e. histo[1] = histogram of channel 1 localisations.
            Note that if considering an image, need to transpose
            the histogram to follow image conventions.
        histo_edges (tuple of lists; each list contains floats):
            Tuple containing (x_edges,y_edges) or (x_edges, y_edges, z_edges)
            where x/y/z_edges are list of floats, each representing the
            edge of the bin in the original space.
            e.g. ([0,10,20],[13,25,20],[2,3,4])
        histo_mask (numpy array): Array containing integers where each should
            represent a different label of the MANUAL segmentation
            0 is reserved for background, is of form [X,Y,Z]
        bin_sizes (tuple of floats): Size of bins of the histogram
            e.g. (23.2, 34.5, 21.3)
        gt_label_map (dict): Dictionary with integer keys
            representing the gt labels for each localisation
            with value being a string, representing the
            real concept e.g. 0:'dog', 1:'cat'
    """

    def __init__(
        self,
        name,
        df,
        dim,
        channels,
        histo={},
        histo_edges=None,
        histo_mask={},
        bin_sizes=None,
        gt_label_map={},
    ):
        """Initialises item"""

        self.name = name
        self.df = df
        self.histo = histo
        self.histo_edges = histo_edges
        self.histo_mask = histo_mask
        self.dim = dim
        self.bin_sizes = bin_sizes
        self.channels = channels
        self.gt_label_map = gt_label_map

    # def save(self, save_loc):
    #     """Save the item

    #         Args:
    #             save_loc (string): Location to save the .pkl file"""

    #     dict = {"name": self.name, "df": self.df, "dim": self.dim,
    #             "channels": self.channels, "histo": self.histo,
    #             "histo_edges": self.histo_edges,
    #             "histo_mask": self.histo_mask,
    #             "bin_sizes": self.bin_sizes}
    #     pickle.dump(dict, open(save_loc, "wb"), pickle.HIGHEST_PROTOCOL)

    # def load(self, input_file):
    #     """ Loads item saved as .pkl file

    #         Args:
    #             input_file (string) : Location of the .pkl file to
    #                 load dataitem from"""

    #     with open(input_file, 'rb') as f:
    #         dict = pickle.load(f)
    #     self.__init__(name=dict['name'], df=dict['df'], dim=dict['dim'],
    #                   channels=dict['channels'],
    #                   histo=dict['histo'],
    #                   histo_edges=dict['histo_edges'],
    #                   histo_mask=dict['histo_mask'],
    #                   bin_sizes=dict['bin_sizes'])

    def coord_2_histo(
        self,
        histo_size,
        cmap=["Greens", "Reds", "Blues", "Purples"],
        vis_interpolation="linear",
    ):
        """Converts localisations into histogram of desired size,
        with option to plot the image (histo.T).
        Note the interpolation is only applied for visualisation,
        not for the actual data in the histogram!

        Args:
            histo_size (tuple): Tuple representing number of
                bins/pixels in x,y,z
            cmap (list of strings) : The colourmaps used to
                plot the histograms
            plot (bool): Whether to plot the output
            vis_interpolation (string): How to inerpolate
                the image for visualisation"""

        # get max and min x/y/(z) values
        df_max = self.df.max()
        df_min = self.df.min()

        if self.dim == 2:
            x_bins, y_bins = histo_size
            x_max = df_max["x"][0]
            y_max = df_max["y"][0]
            x_min = df_min["x"][0]
            y_min = df_min["y"][0]
        elif self.dim == 3:
            x_bins, y_bins, z_bins = histo_size
            z_max = df_max["z"][0]
            z_min = df_min["z"][0]

        # if instead want desired bin size e.g. 50nm, 50nm, 50nm
        # number of bins required for desired bin_size
        # note need to check if do this that agrees with np.digitize
        # and need to make sure that same issue we had before
        # with the last localisation is dealt with
        # x_bins = int((self.max['x'] - self.min['x']) / bin_size[0])
        # y_bins = int((self.max['y'] - self.min['y']) / bin_size[1])
        # z_bins = int((self.max['z'] - self.min['z']) / bin_size[2])

        # size of actual bins, given the number of bins (should be
        # very close to desired tests size)
        x_bin_size = (x_max - x_min) / x_bins
        y_bin_size = (y_max - y_min) / y_bins
        # need to increase bin size very marginally to include last localisation
        x_bin_size = x_bin_size * 1.001
        y_bin_size = y_bin_size * 1.001
        # location of edges of histogram, based on actual tests size
        x_edges = [x_min + x_bin_size * i for i in range(x_bins + 1)]
        y_edges = [y_min + y_bin_size * i for i in range(y_bins + 1)]
        # treat z separately, as often only in 2D
        if self.dim == 3:
            z_bin_size = (z_max - z_min) / z_bins
            # need to increase bin size very marginally to include last localisation
            z_bin_size = z_bin_size * 1.001
            z_edges = [z_min + z_bin_size * i for i in range(z_bins + 1)]

        # size per tests in nm; location of histo edges in original space
        if self.dim == 2:
            self.bin_sizes = (x_bin_size, y_bin_size)
            self.histo_edges = (x_edges, y_edges)
        if self.dim == 3:
            self.bin_sizes = (x_bin_size, y_bin_size, z_bin_size)
            self.histo_edges = (x_edges, y_edges, z_edges)

        print("-- Bin sizes -- ")
        print(self.bin_sizes)

        if self.dim == 2:
            # 2D histogram for every channel, assigned to self.histo (dict)
            for chan in self.channels:
                df = self.df.filter(pl.col("channel") == chan)
                sample = np.array((df["x"], df["y"]))
                sample = np.swapaxes(sample, 0, 1)
                # (D, N) where D is
                # self.dim and N is number of localisations
                self.histo[chan], _ = np.histogramdd(sample, bins=self.histo_edges)

        if self.dim == 3:
            # 3D histogram for every channel, assigned to self.histo (dict)
            for chan in self.channels:
                df = self.df[self.df["channel"] == chan]
                sample = np.array((df["x"], df["y"], df["z"]))
                sample = np.swapaxes(sample, 0, 1)
                # (D, N) where D is self.dim and N is number of
                # localisations
                self.histo[chan], _ = np.histogramdd(sample, bins=self.histo_edges)

        plt.close()

        # work out pixel for each localisations
        self._coord_2_pixel()

        # check digitize agree
        # df_min = self.df.min()
        # x_min = df_min["x_pixel"][0]
        # y_min = df_min["y_pixel"][0]

        # df_max = self.df.max()
        # x_max = df_max["x_pixel"][0]
        # y_max = df_max["y_pixel"][0]

        # print('check')
        # print(x_min, x_max)
        # print(y_min, y_max)

        # print('check')
        # for chan in self.channels:
        #    df = self.df.filter(pl.col("channel") == chan)
        #    my_x = df["x_pixel"].to_numpy()
        #    my_y = df["y_pixel"].to_numpy()
        #    their_x = np.digitize(df['x'], bins=self.histo_edges[0])
        #    their_y = np.digitize(df['y'], bins=self.histo_edges[1])
        #    print('assert equal')
        #    np.testing.assert_array_equal(my_x, their_x-1)
        #    np.testing.assert_array_equal(my_y, their_y-1)

    def _coord_2_pixel(self):
        """Calculate the pixels corresponding to each localisation"""

        # necessary for pd.eval below
        df_min = self.df.min()
        x_min = df_min["x"][0]
        y_min = df_min["y"][0]

        if self.dim == 2:
            x_pixel_width, y_pixel_width = self.bin_sizes
        elif self.dim == 3:
            x_pixel_width, y_pixel_width, z_pixel_width = self.bin_sizes

        # calculate pixel indices for localisations
        self.df = self.df.select(
            [
                pl.all(),
                pl.col("x").map(lambda q: (q - x_min) / x_pixel_width).alias("x_pixel"),
                pl.col("y").map(lambda q: (q - y_min) / y_pixel_width).alias("y_pixel"),
            ]
        )
        # floor the pixel locations
        self.df = self.df.with_column(pl.col("x_pixel").cast(int, strict=True))
        self.df = self.df.with_column(pl.col("y_pixel").cast(int, strict=True))

        # localisations at the end get assigned to outside the histogram,
        # therefore need to be assigned to previous pixel
        # self.df = self.df.with_column(
        #    pl.when(pl.col("x_pixel") == self.df.max()["x_pixel"][0])
        #    .then(self.df.max()["x_pixel"][0] - 1)
        #    .otherwise(pl.col("x_pixel"))
        #    .alias("x_pixel")
        # )
        # self.df = self.df.with_column(
        #    pl.when(pl.col("y_pixel") == self.df.max()["y_pixel"][0])
        #    .then(self.df.max()["y_pixel"][0] - 1)
        #    .otherwise(pl.col("y_pixel"))
        #    .alias("y_pixel")
        # )

        if self.dim == 3:
            z_min = df_min["z"][0]
            # calculate pixel indices for localisations
            self.df = self.df.select(
                [
                    pl.all(),
                    pl.col("z")
                    .map(lambda q: (q - z_min) / z_pixel_width)
                    .alias("z_pixel"),
                ]
            )
            # floor the pixel locations
            self.df = self.df.with_column(pl.col("z_pixel").cast(int, strict=True))
            # localisations at the end get assigned to outside the histogram,
            # therefore need to be assigned
            # to previous pixel
            # self.df = self.df.with_column(
            #    pl.when(pl.col("z_pixel") == self.df.max()["z_pixel"][0])
            #    .then(self.df.max()["z_pixel"][0] - 1)
            #    .otherwise(pl.col("z_pixel"))
            #    .alias("z_pixel")
            # )

    def manual_segment(self, cmap=["green", "red", "blue", "bop purple"]):
        """Manually segment the image (histogram.T). Return the segmented
        histogram and extra column in dataframe corresponding to label.
        0 should be reserved for background

        Args:
            cmap (list of strings) : Colourmaps napari uses to
                plot the histograms
        """

        # if already has gt label raise error
        if "gt_label" in self.df.columns:
            raise ValueError(
                "Manual segment cannot be called on a file which\
                              already has gt labels in it"
            )

        if self.dim == 2:
            # overlay all channels for src
            if len(self.channels) != 1:
                # create the viewer and add each channel (first channel on own,
                # then iterate through others)
                colormap_list = cmap
                # note image shape when plotted: [x, y]
                viewer = napari.view_image(
                    self.histo[self.channels[0]].T,
                    name=f"Channel {self.channels[0]}",
                    rgb=False,
                    blending="additive",
                    colormap=colormap_list[0],
                    gamma=2,
                    contrast_limits=[0, 30],
                )
                for index, chan in enumerate(self.channels[1:]):
                    viewer.add_image(
                        self.histo[chan].T,
                        name=f"Channel {chan}",
                        rgb=False,
                        blending="additive",
                        colormap=colormap_list[index + 1],
                        gamma=2,
                        contrast_limits=[0, 30],
                    )
                napari.run()

            # only one channel
            else:
                img = self.histo[self.channels[0]].T
                # create the viewer and add the image
                viewer = napari.view_image(
                    img,
                    name=f"Channel {self.channels[0]}",
                    rgb=False,
                    gamma=2,
                    contrast_limits=[0, 30],
                )
                napari.run()

            # histogram mask should be assigned to GUI output
            self.histo_mask = viewer.layers["Labels"].data.T

        elif self.dim == 3:
            print("segment 3D image")

        # segment the coordinates
        self._manual_seg_pixel_2_coord()

    def _manual_seg_pixel_2_coord(self):
        """Get the localisations associated with manual annotation.
        Each integer should represent a different label, where 0 is reserved
        for background.
        """

        if self.dim == 2:

            # create dataframe
            flatten_mask = np.ravel(self.histo_mask)
            mesh_grid = np.meshgrid(
                range(self.histo_mask.shape[0]), range(self.histo_mask.shape[1])
            )
            x_pixel = np.ravel(mesh_grid[1])
            y_pixel = np.ravel(mesh_grid[0])
            label = flatten_mask
            data = {"x_pixel": x_pixel, "y_pixel": y_pixel, "gt_label": label}
            mask_df = pl.DataFrame(
                data,
                columns=[
                    ("x_pixel", pl.Int64),
                    ("y_pixel", pl.Int64),
                    ("gt_label", pl.Float64),
                ],
            ).sort(["x_pixel", "y_pixel"])

            # join mask dataframe
            self.df = self.df.join(mask_df, how="inner", on=["x_pixel", "y_pixel"])

            # sanity check
            # print(len(self.df))
            # print(self.df.columns)
            # print(self.df.head(10))

        elif self.dim == 3:
            print("segment the 3d coords")

    def mask_pixel_2_coord(self, img_mask: np.ndarray) -> pl.DataFrame:
        """For a given mask over the image (value at each pixel
        normally representing a label), return the dataframe with a column
        giving the value for each localisation. Note that it is
        assumed that the img_mask is a mask of the image,
        therefore have to transpose img_mask for it to be in the same
        configuration as the histogram

        Note we also use this for  labels and when
        the img_mask represents probabilities.

        Args:
            img_mask (np.ndarray): Mask over the image -
            to reiterate, to convert this to histogram space need
            to transpose it

        Returns:
            df (polars dataframe): Original dataframe with
            additional column with the predicted label"""

        if self.dim == 2:
            # list of mask dataframes, each mask dataframe contains
            # (x,y,label) columns
            # transpose the image mask to histogram space
            histo_mask = img_mask.T

            # create dataframe
            flatten_mask = np.ravel(histo_mask)
            mesh_grid = np.meshgrid(
                range(histo_mask.shape[0]), range(histo_mask.shape[1])
            )
            x_pixel = np.ravel(mesh_grid[1])
            y_pixel = np.ravel(mesh_grid[0])
            label = flatten_mask
            data = {"x_pixel": x_pixel, "y_pixel": y_pixel, "pred_label": label}
            mask_df = pl.DataFrame(
                data,
                columns=[
                    ("x_pixel", pl.Int64),
                    ("y_pixel", pl.Int64),
                    ("pred_label", pl.Float64),
                ],
            ).sort(["x_pixel", "y_pixel"])

            # sanity check
            # print(len(self.df))
            # print(self.df.columns)
            # print(self.df.head(10))

            # join mask dataframe
            df = self.df.join(mask_df, how="inner", on=["x_pixel", "y_pixel"])

            # sanity check
            # print(len(df))
            # print(df.columns)
            # print(df.head(10))

            return df

        elif self.dim == 3:
            print("segment the 3d coords")

    def save_df_to_csv(self, csv_loc, drop_zero_label=False, drop_pixel_col=True):
        """Save the dataframe to a .csv with option to drop positions which
           are background and can drop the column containing pixel
           information

        Args:
            csv_loc (String): Save the csv to this location
            drop_zero_label (bool): If True then only non zero
                label positions are saved to csv
            drop_pixel_col (bool): If True then don't save
                the column with x,y,z pixel

        Returns:
            None"""

        save_df = self.df

        if drop_pixel_col:
            # don't want to save x,y pixel to csv
            save_df = save_df.drop("x_pixel")
            save_df = save_df.drop("y_pixel")
            if self.dim == 3:
                save_df = save_df.drop("z_pixel")

        # rearrange so x,y,z, ...,labels,channels
        # if self.dim == 2:
        #    cols = ['x', 'y', 'gt_label', 'channel']
        # elif self.dim == 3:
        #    cols = ['x', 'y', 'z', 'gt_label', 'channel']
        # save_df_cols = save_df.columns
        # cols = [col for col in cols if col in save_df_cols] +
        # [col for col in save_df_cols if col #]not in cols]
        # save_df = save_df[cols]

        # drop rows with zero label
        if drop_zero_label:
            save_df = save_df.filter(pl.col("gt_label") != 0)

        # save to location
        save_df.write_csv(csv_loc, sep=",")

    def save_to_parquet(
        self,
        save_folder,
        drop_zero_label=False,
        drop_pixel_col=False,
        gt_label_map=None,
    ):
        """Save the dataframe to a parquet with option to drop positions which
           are background and can drop the column containing pixel
           information

        Args:
            save_folder (String): Save the df to this folder
            drop_zero_label (bool): If True then only non zero
                label positions are saved to parquet
            drop_pixel_col (bool): If True then don't save
                the column with x,y,z pixel
            gt_label_map (dict): Dictionary with integer keys
                representing the gt labels for each localisation
                with value being a string, representing the
                real concept e.g. 0:'dog', 1:'cat'

        Returns:
            None
        """

        save_df = self.df

        if drop_pixel_col:
            # don't want to save x,y pixel to csv
            save_df = save_df.drop("x_pixel")
            save_df = save_df.drop("y_pixel")
            if self.dim == 3:
                save_df = save_df.drop("z_pixel")

        # drop rows with zero label
        if drop_zero_label:
            save_df = save_df.filter(pl.col("gt_label") != 0)

        # convert to arrow + add in metadata if doesn't exist
        arrow_table = save_df.to_arrow()

        # convert gt label map to bytes
        old_metadata = arrow_table.schema.metadata

        # convert to bytes
        gt_label_map = json.dumps(gt_label_map).encode("utf-8")
        meta_data = {
            "name": self.name,
            "dim": str(self.dim),
            "channels": str(self.channels),
            "gt_label_map": gt_label_map,
            "bin_sizes": str(self.bin_sizes),
        }

        # add in label mapping
        # meta_data.update(gt_label_map)
        # merge existing with new meta data
        merged_metadata = {**meta_data, **(old_metadata or {})}
        arrow_table = arrow_table.replace_schema_metadata(merged_metadata)
        save_loc = os.path.join(save_folder, self.name + ".parquet")
        pq.write_table(arrow_table, save_loc)

        # To access metadata write
        # parquet_table = pq.read_table(file_path)
        # parquet_table.schema.metadata ==> metadata
        # note if accessing keys need
        # parquet_table.schema.metadata[b'key_name'])
        # note that b is bytes

    def load_from_parquet(self, input_file):
        """Loads item saved as .parquet file

        Args:
            input_file (string) : Location of the .parquet file to
                load dataitem from"""

        # read in parquet file
        arrow_table = pq.read_table(input_file)

        # print("loaded metadata", arrow_table.schema.metadata)

        # metadata
        name = arrow_table.schema.metadata[b"name"].decode("utf-8")
        gt_label_map = json.loads(
            arrow_table.schema.metadata[b"gt_label_map"].decode("utf-8")
        )
        if gt_label_map is not None:
            # convert string keys to int keys for the mapping
            gt_label_map = {int(key): value for key, value in gt_label_map.items()}
        # convert string keys to int keys for the mapping
        dim = arrow_table.schema.metadata[b"dim"]
        channels = arrow_table.schema.metadata[b"channels"]
        dim = int(dim)
        channels = ast.literal_eval(channels.decode("utf-8"))
        bin_sizes = arrow_table.schema.metadata[b"bin_sizes"]
        bin_sizes = ast.literal_eval(bin_sizes.decode("utf-8"))
        df = pl.from_arrow(arrow_table)

        self.__init__(
            name=name,
            df=df,
            dim=dim,
            channels=channels,
            gt_label_map=gt_label_map,
            bin_sizes=bin_sizes,
        )

    def get_img_dict(self):
        """Return dictionary of images,
        where each key represents a channel"""

        img_dict = {}
        for key, value in self.histo.items():
            img_dict[key] = value.T

        return img_dict

    def render_histo(self):
        """Render the histogram from the .parquet file
        
        Assumes localisations have associated x_pixel and y_pixel already.

        Args:
            None

        Returns:
            histo (np.histogram) : Histogram of the localisation data
            axis_2_chan (list) : List where the first value is the
                channel in the first axis of the histogram, second value
                is the channel in the second axis of the histogram etc.
                e.g. [1,3] : 1st channel is in 1st axis, 3rd channel in 2nd axis
        """

        histos = []
        axis_2_chan = []

        df_max = self.df.max()
        x_bins = df_max["x_pixel"][0] + 1
        y_bins = df_max["y_pixel"][0] + 1

        for chan in self.channels:
            df = self.df.filter(pl.col("channel") == chan)

            histo = np.empty((x_bins, y_bins))
            df = df.groupby(by=["x_pixel", "y_pixel"]).count()
            x_pixels = df["x_pixel"].to_numpy()
            y_pixels = df["y_pixel"].to_numpy()
            counts = df["count"].to_numpy()
            histo[x_pixels, y_pixels] = counts

            histos.append(histo)
            axis_2_chan.append(chan)

        histo = np.stack(histos)

        return histo, axis_2_chan

    def render_seg(self):
        """Render the segmentation of the histogram"""

        labels = self.df.select(pl.col("gt_label")).to_numpy()
        x_pixels = self.df.select(pl.col("x_pixel")).to_numpy()
        y_pixels = self.df.select(pl.col("y_pixel")).to_numpy()

        histo_width = np.max(x_pixels) + 1
        histo_height = np.max(y_pixels) + 1

        histo = np.empty((histo_width, histo_height))

        histo[x_pixels, y_pixels] = labels

        return histo



#!/usr/bin/env python
"""Classic segmentation module

Take in items and segment using classic methods
"""

import yaml
import os
from locpix.preprocessing import datastruc
from locpix.visualise import vis_img
from locpix.img_processing import watershed
import numpy as np
import pickle as pkl
import argparse
from locpix.scripts.img_seg import classic_config
import tkinter as tk
from tkinter import filedialog


def main():

    parser = argparse.ArgumentParser(description="Classic")
    parser.add_argument(
        "-i",
        "--project_directory",
        action="store",
        type=str,
        help="the location of the project directory",
    )
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        type=str,
        help="the location of the .yaml configuaration file\
                             for preprocessing",
    )

    args = parser.parse_args()

    # input project directory
    if args.project_directory is not None:
        project_folder = args.project_directory
    else:
        root = tk.Tk()
        root.withdraw()
        project_folder = filedialog.askdirectory(title="Project directory")

    # configuration folder
    if args.config is not None:
        # load yaml
        with open(args.config, "r") as ymlfile:
            config = yaml.safe_load(ymlfile)
            classic_config.parse_config(config)
    else:
        config = classic_config.config_gui()

    # list items
    input_folder = os.path.join(project_folder, "annotate/annotated")
    try:
        files = os.listdir(input_folder)
    except FileNotFoundError:
        raise ValueError("There should be some files to open")

    # if output directory not present create it
    output_membrane_prob = os.path.join(project_folder, "classic/membrane/prob_map")
    if os.path.exists(output_membrane_prob):
            raise ValueError(f"Cannot proceed as {output_membrane_prob} already exists")
    else:
        os.makedirs(output_membrane_prob)

    # if output directory not present create it
    output_cell_df = os.path.join(project_folder, "classic/cell/seg_dataframes")
    if os.path.exists(output_cell_df):
            raise ValueError(f"Cannot proceed as {output_cell_df} already exists")
    else:
        os.makedirs(output_cell_df)

    # if output directory not present create it
    output_cell_img = os.path.join(project_folder, "classic/cell/seg_img")
    if os.path.exists(output_cell_img):
            raise ValueError(f"Cannot proceed as {output_cell_img} already exists")
    else:
        os.makedirs(output_cell_img)

    for file in files:
        item = datastruc.item(None, None, None, None)
        item.load_from_parquet(os.path.join(input_folder, file))

        print("bin sizes", item.bin_sizes)

        # load in histograms
        input_histo_folder = os.path.join(project_folder, "annotate/histos")
        histo_loc = os.path.join(input_histo_folder, item.name + ".pkl")
        with open(histo_loc, "rb") as f:
            histo = pkl.load(f)

        # ---- segment membranes ----
        if config["sum_chan"] is False:
            img = histo[0].T  # consider only the zero channel
        elif config["sum_chan"] is True:
            img = histo[0].T + histo[1].T
        else:
            raise ValueError("sum_chan should be true or false")
        log_img = vis_img.manual_threshold(
            img, config["vis_threshold"], how=config["vis_interpolate"]
        )
        grey_log_img = vis_img.img_2_grey(log_img)  # convert img to grey
        grey_img = vis_img.img_2_grey(img)  # convert img to grey

        # img mask
        semantic_mask = (grey_log_img - np.min(grey_log_img)) / (
            np.max(grey_log_img) - np.min(grey_log_img)
        )

        # ---- segment cells ----
        # get markers
        markers_loc = os.path.join(project_folder, "markers")
        markers_loc = os.path.join(markers_loc, item.name + ".npy")
        try:
            markers = np.load(markers_loc)
        except FileNotFoundError:
            raise ValueError(
                "Couldn't open the file/No markers were found in relevant location"
            )

        # tested very small amount annd line below is better than doing
        # watershed on grey_log_img
        instance_mask = watershed.watershed_segment(
            grey_img, coords=markers
        )  # watershed on the grey image

        # ---- save ----

        # save membrane mask
        output_membrane_prob = os.path.join(project_folder, "classic/membrane/prob_map")
        save_loc = os.path.join(output_membrane_prob, item.name + ".npy")
        np.save(save_loc, semantic_mask)

        # save markers
        np.save(markers_loc, markers)

        # save instance mask to dataframe
        df = item.mask_pixel_2_coord(instance_mask)
        item.df = df
        output_cell_df = os.path.join(project_folder, "classic/cell/seg_dataframes")
        item.save_to_parquet(output_cell_df, drop_zero_label=False, drop_pixel_col=True)

        imgs = {key: value.T for key, value in histo.items()}

        # save cell segmentation image - consider only zero channel
        output_cell_img = os.path.join(project_folder, "classic/cell/seg_img")
        save_loc = os.path.join(output_cell_img, item.name + ".png")
        vis_img.visualise_seg(
            imgs,
            instance_mask,
            item.bin_sizes,
            channels=[0],
            threshold=config["vis_threshold"],
            how=config["vis_interpolate"],
            origin="upper",
            blend_overlays=True,
            alpha_seg=0.5,
            save=True,
            save_loc=save_loc,
            four_colour=True,
        )

        # save yaml file
        yaml_save_loc = os.path.join(project_folder, "classic.yaml")
        with open(yaml_save_loc, "w") as outfile:
            yaml.dump(config, outfile)


if __name__ == "__main__":
    main()



""" Visualise module for images

Contains functions for visualising histograms
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap, ListedColormap
import networkx as nx
from skimage.future.graph import RAG
from skimage.transform import resize

_interpolate = {
    "log2": lambda d: np.log2(d),
    "log10": lambda d: np.log10(d),
    "linear": lambda d: d,
}


def manual_threshold(array, threshold, how="linear") -> np.ndarray:
    """Perform manual threshold of the input array

    Args:
        array (np.ndarray): numpy array
        threshold (float or int): Threshold above
            which array will be thresholded
        how (string): How to interpolate the array

    Returns:
        thresh_array (np.ndarray): The thresholded output on the input array"""

    # interpolate the image
    array = _interpolate[how](array)
    array = np.clip(array, 0, None)  # clip to at least 0
    return np.where(array > threshold, array, 0)


def label_2_4_colours(labels: np.ndarray):
    """This takes in an array of labels
    e.g. [5,0,2,2,17,1,...]
    and returns a list where each label is now
    in range 0-3 such that no 2 integers next to
    each other are the same.
    This can be used for to plot so no 2 colours
    next to each other are the same

    Args:
        labels (np.ndarray): Array of ints
        each representing a unique label

    Returns:
        Array of ints each representing a
        unique label (between 0 and 3) no
        2 of same label should be next to each other"""

    graph = RAG(labels)
    d = nx.coloring.greedy_color(graph)
    return np.vectorize(d.get)(labels)


def img_2_grey(img: np.ndarray) -> np.ndarray:
    """converts img to 0 255 greyscale image

    Args:
        img (np.ndarray) : Image to be converted to greyscale

    Returns:
        img (np.ndarray): Image in uint8 form greyscale"""

    mini = np.min(img)
    maxi = np.max(img)
    img = ((img - mini) / (maxi - mini)) * 255
    return img.astype("uint8")


def visualise_seg(
    image_dict,
    segmentation,
    bin_sizes,
    channels,
    threshold=0,
    how="linear",
    alphas=[1, 0.5, 0.2, 0.1],
    alpha_seg=0.8,
    blend_overlays=False,
    cmap_img=None,
    cmap_seg=["m", "g", "lightsalmon", "r", "b"],
    figsize=(10, 10),
    origin="upper",
    save=False,
    save_loc=None,
    four_colour=True,
    background_one_colour=False,
):
    """Take in image and the associated segmentation and plot it,
    with option to convert to 4 colours and also option to choose to
    plot background as one particular colour
    (only relevant if four colour is true)

    Args:
        image_dict (dict) : Dictionary where each key is
            the channel for the respective image (which is
            the transpose of the histogram) i.e. img_dict[0]
            should be a img (histogram.T) for the 0th channel
        segmentation (np.ndarray) : Array of integers where
            each represents unique label in the segmentation
        bin_sizes (list) : List of sizes of the bins of the
            histogram
        channels (list): List of channels to plot
        threshold (int) : Threshold applied to images
            when plotting
        how (string) : Interpolation applied to image
            when plotting
        alphas (list) : List of alpha to be used in plt.imshow
        alpha_seg (float) : Value of alpha for segmentation
        blend_overlays (bool) : Whether to blend img and seg
        cmap_img (list) : List of cmaps to plot images
            (transpose of histograms)
        cmap_seg (list) : List of cmaps to plot segmentations
        figsize (tuple) : Tuple of figure size in 2D (x,y) sizes
        origin (string) : Location of origin in plot, image
            convention is 'upper' but for cartesian use 'lower'
        save (bool) : Whether to save
        save_loc (string) : Where to save if chosen to save
        four_colour (bool) : Whether to convert localisations
            to 4 colour
        background_one_colour (bool) : Whether to keep background
            as all same colour when doing 4 colour conversion

    """

    if cmap_img is None:
        cmap_img = [
            LinearSegmentedColormap.from_list("", ["k", "c"], N=256),
            LinearSegmentedColormap.from_list("", ["k", "y"], N=256),
            LinearSegmentedColormap.from_list("", ["k", "b"], N=256),
            LinearSegmentedColormap.from_list("", ["k", "g"], N=256),
        ]
    else:
        cmap_img = [
            LinearSegmentedColormap.from_list("", ["k", i], N=256) for i in cmap_img
        ]

    if four_colour:
        if background_one_colour:
            four_colour_seg = label_2_4_colours(segmentation)
            # now set all regions which were 0 before to 0
            # and increase four colour by 1
            # and add white to cmap_seg --> ensures
            # background comes up as untouched
            segmentation = np.where(segmentation != 0, four_colour_seg + 1, 0)
            cmap_seg = ["w"] + cmap_seg
        else:
            segmentation = label_2_4_colours(segmentation)

    fig, ax = plt.subplots(figsize=figsize)
    cmap_seg = ListedColormap(cmap_seg)

    # patches = []  # legend creation
    # plot for each channel in sequence
    for index, chan in enumerate(channels):
        img = manual_threshold(image_dict[chan], threshold=threshold, how=how)
        img = img_2_grey(img)
        img = np.where(img > 100, 255, 0)

        # resize img according to bin size - so that pixel size same in x and y
        h_old, w_old = img.shape
        img = resize(img, (h_old * bin_sizes[1] / bin_sizes[0], w_old))
        segmentation = resize(
            segmentation, (h_old * bin_sizes[1] / bin_sizes[0], w_old)
        )
        img = img[0:h_old, 0:w_old]
        segmentation = segmentation[0:h_old, 0:w_old]

        ax.imshow(
            img, cmap=cmap_img[index], origin=origin, alpha=alphas[index]
        )  # cmap=cmap_list[index]
        # x,y scale bars
        x = [15, 115]
        y = [465, 465]
        ax.plot(x, y, color="w", lw=3)
        ax.text(
            35,
            490,
            f"{bin_sizes[0]*100/1000:.0f}\u03BCm",
            fontsize=25,
            color="w",
            bbox={"facecolor": "black", "edgecolor": "none", "pad": 2, "alpha": 0.7},
        )
        ax.set_axis_off()
        # cmap = plt.cm.get_cmap(cmap_list[index]) # legend creation

        # legend creation
        # cmap = plt.cm.get_cmap(cmap_img[index])
        # patches.append(mpatches.Patch(color=cmap(255), label=f"Chan {chan}"))
        # plt.legend(
        #    handles=patches,
        #    bbox_to_anchor=(0.8, 1),
        #    loc=2,
        #    borderaxespad=0.0,
        #    prop={"size": 15},
        # )

    if blend_overlays:
        alphas = alpha_seg
    elif blend_overlays is False:
        alphas = np.where(segmentation > 0, alpha_seg, 0).astype(float)
    plt.imshow(segmentation, cmap=cmap_seg, alpha=alphas, origin=origin)

    if save:
        if save_loc is None:
            raise ValueError("Need a save location!")
        plt.savefig(save_loc, bbox_inches="tight")

    plt.close()



"""Annotate config

Module which defines functions for managing the config
files for annotate.
Including:

Creating a GUI for user to choose configuration
Parsing config file to check all specified"""

from PyQt5.QtWidgets import (
    QLabel,
    QApplication,
    QWidget,
    QHBoxLayout,
    QMessageBox,
    QLineEdit,
    QFormLayout,
    QCheckBox,
    QListWidget,
    QPushButton,
    QFileDialog,
)
from PyQt5.QtGui import QIntValidator, QDoubleValidator
from PyQt5.QtCore import Qt
import yaml

default_config_keys = [
    "x_bins",
    "y_bins",
    "z_bins",
    "dim",
    "plot",
    "vis_interpolation",
    "drop_zero_label",
    "gt_label_map",
    "save_img",
    "save_threshold",
    "save_interpolate",
    "background_one_colour",
    "four_colour",
    "alphas",
    "alpha_seg",
    "cmap_seg",
    "fig_size",
    "vis_channels",
]


class InputWidget(QWidget):
    """Input
    Widget to take in user configuration
    Child of QWidget.
    Args:
        None
    Attributes:

    """

    def __init__(self, config):
        """Constructor
        Args:
            files (list) : List of files that will be preprocessed
            config (dict) : Dictionary containing the configuration
        """

        super().__init__()
        self.flo = QFormLayout()

        # Load .yaml with button
        self.load_button = QPushButton("Load yaml")
        self.load_button.clicked.connect(self.load_yaml)
        self.flo.addRow(self.load_button)

        self.x_bins = QLineEdit("500")
        self.x_bins.setValidator(QIntValidator())
        self.x_bins.setToolTip("Number of bins in x dimension")
        self.flo.addRow("X bins", self.x_bins)

        self.y_bins = QLineEdit("500")
        self.y_bins.setValidator(QIntValidator())
        self.y_bins.setToolTip("Number of bins in y dimension")
        self.flo.addRow("Y bins", self.y_bins)

        self.z_bins = QLineEdit("500")
        self.z_bins.setValidator(QIntValidator())
        self.z_bins.setToolTip("Number of bins in z dimension")
        self.flo.addRow("Z bins", self.z_bins)

        self.dim = QLineEdit("2")
        self.dim.setValidator(QIntValidator())
        self.dim.setToolTip("Dimensions of the data either 2 or 3")
        self.flo.addRow("Dimensions", self.dim)

        self.plot = QCheckBox()
        self.plot.setChecked(True)
        self.plot.setToolTip("Whether plot should occur to screen")
        self.flo.addRow("Plot", self.plot)

        self.vis_interpolation = QListWidget()
        self.vis_interpolation.insertItem(0, "log2")
        self.vis_interpolation.insertItem(1, "log10")
        self.vis_interpolation.insertItem(2, "linear")
        self.vis_interpolation.item(0).setSelected(True)
        self.vis_interpolation.setToolTip(
            "Interpolation applied to the histogram when visualising"
            "the image of the histogram"
        )
        self.flo.addRow("Vis interpolation", self.vis_interpolation)

        self.drop_zero_label = QCheckBox()
        help = (
            "When saving annotations - you can choose to"
            "not save the localisations associated with the background"
            "which are assigned a label of zero"
        )
        self.drop_zero_label.setToolTip(help)
        self.flo.addRow("Drop zero label", self.drop_zero_label)

        h_box = QHBoxLayout()
        zero_label = QLabel("0: ")
        self.gt_label_map_zero = QLineEdit("background")
        h_box.addWidget(zero_label)
        h_box.addWidget(self.gt_label_map_zero)
        one_label = QLabel("1: ")
        self.gt_label_map_one = QLineEdit("membrane")
        h_box.addWidget(one_label)
        h_box.addWidget(self.gt_label_map_one)
        self.gt_label_map_zero.setToolTip("0 label in real terms")
        self.gt_label_map_one.setToolTip("1 lable in real terms")
        self.flo.addRow("Gt label map", h_box)

        self.save_img = QCheckBox()
        self.save_img.setChecked(True)
        self.save_img.setToolTip("Save segmentation images settings")
        self.flo.addRow("Save image", self.save_img)

        self.save_threshold = QLineEdit("0")
        self.save_threshold.setValidator(QDoubleValidator())
        self.flo.addRow("Save threshold", self.save_threshold)

        self.save_interpolation = QListWidget()
        self.save_interpolation.insertItem(0, "log2")
        self.save_interpolation.insertItem(1, "log10")
        self.save_interpolation.insertItem(2, "linear")
        self.save_interpolation.item(0).setSelected(True)
        self.flo.addRow("Save interpolation", self.save_interpolation)

        self.background_one_colour = QCheckBox()
        self.background_one_colour.setChecked(True)
        self.flo.addRow("Background one colour", self.background_one_colour)

        self.four_colour = QCheckBox()
        self.four_colour.setChecked(True)
        self.flo.addRow("Four colour", self.four_colour)

        h_box = QHBoxLayout()
        zero_alpha = QLabel("Alpha chan zero: ")
        self.alpha_zero = QLineEdit("1")
        self.alpha_zero.setValidator(QDoubleValidator())
        h_box.addWidget(zero_alpha)
        h_box.addWidget(self.alpha_zero)
        one_alpha = QLabel("Alpha chan one: ")
        self.alpha_one = QLineEdit(".5")
        self.alpha_one.setValidator(QDoubleValidator())
        h_box.addWidget(one_alpha)
        h_box.addWidget(self.alpha_one)
        two_alpha = QLabel("Alpha chan two: ")
        self.alpha_two = QLineEdit(".2")
        self.alpha_two.setValidator(QDoubleValidator())
        h_box.addWidget(two_alpha)
        h_box.addWidget(self.alpha_two)
        three_alpha = QLabel("Alpha chan three: ")
        self.alpha_three = QLineEdit(".1")
        self.alpha_three.setValidator(QDoubleValidator())
        h_box.addWidget(three_alpha)
        h_box.addWidget(self.alpha_three)
        self.flo.addRow("Alphas", h_box)

        self.alpha_seg = QLineEdit("0.8")
        self.alpha_seg.setValidator(QDoubleValidator())
        self.flo.addRow("Alpha seg", self.alpha_seg)

        h_box = QHBoxLayout()
        zero_cmap = QLabel("Zero cmap: ")
        self.zero_cmap = QLineEdit("k")
        h_box.addWidget(zero_cmap)
        h_box.addWidget(self.zero_cmap)
        one_cmap = QLabel("One cmap: ")
        self.one_cmap = QLineEdit("y")
        h_box.addWidget(one_cmap)
        h_box.addWidget(self.one_cmap)
        self.flo.addRow("Cmap seg", h_box)

        h_box = QHBoxLayout()
        fig_size_x = QLabel("Fig size x: ")
        self.fig_size_x = QLineEdit("10")
        self.fig_size_x.setValidator(QIntValidator())
        h_box.addWidget(fig_size_x)
        h_box.addWidget(self.fig_size_x)
        fig_size_y = QLabel("Fig size y: ")
        self.fig_size_y = QLineEdit("10")
        self.fig_size_y.setValidator(QIntValidator())
        h_box.addWidget(fig_size_y)
        h_box.addWidget(self.fig_size_y)
        self.flo.addRow("Fig size", h_box)

        self.vis_channels = QListWidget()
        self.vis_channels.insertItem(0, "0")
        self.vis_channels.insertItem(1, "1")
        self.vis_channels.insertItem(2, "2")
        self.vis_channels.insertItem(3, "3")
        self.vis_channels.setSelectionMode(2)
        self.vis_channels.setToolTip(
            "Choice of which channels user wants to view for seg"
        )
        self.vis_channels.item(0).setSelected(True)
        self.flo.addRow("Channels", self.vis_channels)

        self.setLayout(self.flo)

        self.config = config

    def load_yaml(self):
        """Load the yaml"""

        # Load yaml
        fname = QFileDialog.getOpenFileName(
            self, "Open file", "/home/some/folder", "Yaml (*.yaml)"
        )

        fname = str(fname[0])
        if fname != "":
            with open(fname, "r") as ymlfile:
                load_config = yaml.safe_load(ymlfile)
                if sorted(load_config.keys()) == sorted(default_config_keys):
                    self.load_config(load_config)
                else:
                    print("Can't load in as keys don't match!")

    def load_config(self, load_config):
        """Load the config into the gui

        Args:
            load_config (yaml file): Config file
                to load into the gui"""

        self.x_bins.setText(str(load_config["x_bins"]))
        self.y_bins.setText(str(load_config["y_bins"]))
        self.z_bins.setText(str(load_config["z_bins"]))
        self.dim.setText(str(load_config["dim"]))
        self.plot.setCheckState(load_config["plot"])
        self.vis_interpolation.clearSelection()
        item = self.vis_interpolation.findItems(
            load_config["vis_interpolation"], Qt.MatchFlag.MatchExactly
        )
        item[0].setSelected(True)
        self.drop_zero_label.setCheckState(load_config["drop_zero_label"])
        self.gt_label_map_zero.setText(str(load_config["gt_label_map"][0]))
        self.gt_label_map_one.setText(str(load_config["gt_label_map"][1]))
        self.save_img.setCheckState(load_config["save_img"])
        self.save_threshold.setText(str(load_config["save_threshold"]))
        self.save_interpolation.clearSelection()
        item = self.save_interpolation.findItems(
            load_config["save_interpolate"], Qt.MatchFlag.MatchExactly
        )
        item[0].setSelected(True)
        self.background_one_colour.setCheckState(load_config["background_one_colour"])
        self.four_colour.setCheckState(load_config["four_colour"])
        alphas = load_config["alphas"]
        self.alpha_zero.setText(str(alphas[0]))
        self.alpha_one.setText(str(alphas[1]))
        self.alpha_two.setText(str(alphas[2]))
        self.alpha_three.setText(str(alphas[3]))
        self.alpha_seg.setText(str(load_config["alpha_seg"]))
        self.zero_cmap.setText(load_config["cmap_seg"][0])
        self.one_cmap.setText(load_config["cmap_seg"][1])
        self.fig_size_x.setText(str(load_config["fig_size"][0]))
        self.fig_size_y.setText(str(load_config["fig_size"][1]))
        self.vis_channels.clearSelection()
        for chan in load_config["vis_channels"]:
            item = self.vis_channels.findItems(str(chan), Qt.MatchFlag.MatchExactly)
            item[0].setSelected(True)

    def set_config(self, config):
        """Set the configuration file

        Args:
            config (dictionary) : Configuration dict"""

        config["x_bins"] = int(self.x_bins.text())
        config["y_bins"] = int(self.y_bins.text())
        config["z_bins"] = int(self.z_bins.text())
        config["dim"] = int(self.dim.text())
        config["plot"] = self.plot.isChecked()
        config["vis_interpolation"] = self.vis_interpolation.selectedItems()[0].text()
        config["drop_zero_label"] = self.drop_zero_label.isChecked()
        config["gt_label_map"] = {
            0: self.gt_label_map_zero.text(),
            1: self.gt_label_map_one.text(),
        }
        config["save_img"] = self.save_img.isChecked()
        config["save_threshold"] = float(self.save_threshold.text())
        config["save_interpolate"] = self.save_interpolation.selectedItems()[0].text()
        config["background_one_colour"] = self.background_one_colour.isChecked()
        config["four_colour"] = self.four_colour.isChecked()
        config["alphas"] = [
            float(self.alpha_zero.text()),
            float(self.alpha_one.text()),
            float(self.alpha_two.text()),
            float(self.alpha_three.text()),
        ]
        config["alpha_seg"] = float(self.alpha_seg.text())
        config["cmap_seg"] = [self.zero_cmap.text(), self.one_cmap.text()]
        config["fig_size"] = [int(self.fig_size_x.text()), int(self.fig_size_y.text())]
        config["vis_channels"] = [
            int(item.text()) for item in self.vis_channels.selectedItems()
        ]

        # check config is correct
        parse_config(config)

    def closeEvent(self, event):
        """When closing the widget this function will be overloaded.
            It will ask the user if they are sure, on closing the coords of the markers
                will be extracted
        Args:
            event (QEvent): Event triggered by click"""

        # Confirm user choice
        reply = QMessageBox()
        reply.setText("Are you sure you want to close? (Config is saved on closing)")
        reply.setStandardButtons(
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
        )
        reply = reply.exec()
        if reply == QMessageBox.StandardButton.Yes:
            self.set_config(self.config)
            event.accept()
        else:
            event.ignore()


def config_gui():
    """Config gui
    This function opens up a GUI for user to specify
    the configuration
    List of files files can then be unchecked if users want to ignore

    Attributes:
        None"""

    app = QApplication([])  # sys.argv if need command line inputs
    # create widget
    config = {}
    widget = InputWidget(config)
    widget.show()
    app.exec()
    return config


def parse_config(config):
    """Parse config
    This function takes in the configuration .yaml
    file and checks all the necessary arguments are
    specified

    Attributes:
        config (yml file): The configuration
            .yaml file"""

    if sorted(config.keys()) != sorted(default_config_keys):
        raise ValueError(
            "Did not specify necessary default \
            configutation arguments"
        )



#!/usr/bin/env python
"""Get markers from image
"""

import yaml
import os
from locpix.preprocessing import datastruc
from locpix.visualise import vis_img
from locpix.img_processing import watershed
import numpy as np
import pickle as pkl
import argparse
from locpix.scripts.img_seg import get_markers_config
import tkinter as tk
from tkinter import filedialog


def main():

    parser = argparse.ArgumentParser(description="Get markers")
    # config_group = parser.add_mutually_exclusive_group(required=True)
    parser.add_argument(
        "-i",
        "--project_directory",
        action="store",
        type=str,
        help="the location of the project directory",
    )
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        type=str,
        help="the location of the .yaml configuaration file\
                             for get markers",
    )

    args = parser.parse_args()

    # input project directory
    if args.project_directory is not None:
        project_folder = args.project_directory
    else:
        root = tk.Tk()
        root.withdraw()
        project_folder = filedialog.askdirectory(title="Project directory")

    if args.config is not None:
        # load yaml
        with open(args.config, "r") as ymlfile:
            config = yaml.safe_load(ymlfile)
            get_markers_config.parse_config(config)
    else:
        config = get_markers_config.config_gui()

    # list items
    input_folder = os.path.join(project_folder, "preprocess/no_gt_label")
    try:
        files = os.listdir(input_folder)
    except FileNotFoundError:
        raise ValueError("There should be some files to open")

    # if output directory not present create it
    markers_folder = os.path.join(project_folder, "markers")
    if not os.path.exists(markers_folder):
        print("Making folder")
        os.makedirs(markers_folder)
    else:
        raise ValueError(
            "Will not get markers\
                         as there is already files in folder"
        )

    for file in files:
        item = datastruc.item(None, None, None, None)
        item.load_from_parquet(os.path.join(input_folder, file))

        # load in histograms
        input_histo_folder = os.path.join(project_folder, "annotate/histos")
        histo_loc = os.path.join(input_histo_folder, item.name + ".pkl")
        with open(histo_loc, "rb") as f:
            histo = pkl.load(f)

        # convert image to more visible form
        img = histo[0].T  # consider only the zero channel
        log_img = vis_img.manual_threshold(
            img, config["vis_threshold"], how=config["vis_interpolate"]
        )
        grey_log_img = vis_img.img_2_grey(log_img)  # convert img to grey

        markers = watershed.get_markers(grey_log_img)

        markers_loc = os.path.join(markers_folder, item.name + ".npy")

        # save
        np.save(markers_loc, markers)

    # save yaml file
    yaml_save_loc = os.path.join(project_folder, "get_markers.yaml")
    with open(yaml_save_loc, "w") as outfile:
        yaml.dump(config, outfile)


if __name__ == "__main__":
    main()



"""Preprocessing module.

This module contains functions to preprocess the data,
including (add as added):
- convert .csv to datastructure
"""

import polars as pl
import os
from . import datastruc


def csv_to_datastruc(
    input_file, dim, channel_col, frame_col, x_col, y_col, z_col, channel_choice=None
):
    """Loads in .csv and converts to the required datastructure.
    
    Currently considers the following columns: channel frame x y z
    Also user can specify the channels they want to consider, these
    should be present in the channels column

    Args:
        input_file (string) : Location of the .csv
        save_loc (string) : Location to save datastructure to
        dim (int) : Dimensions to consider either 2 or 3
        channel_col (string) : Name of column which gives channel
            for localisation
        frame_col (string) : Name of column which gives frame for localisation
        x_col (string) : Name of column which gives x for localisation
        y_col (string) : Name of column which gives y for localisation
        z_col (string) : Name of column which gives z for localisation
        channel_choice (list of ints) : If specified then this will be list
            of integers representing channels to be considered

    Returns:
        datastruc (SMLM_datastruc) : Datastructure containg the data
    """

    # Check dimensions correctly specified
    if dim != 2 and dim != 3:
        raise ValueError("Dimensions must be 2 or 3")
    if dim == 2 and z_col:
        raise ValueError("If dimensions are two no z should be specified")
    if dim == 3 and not z_col:
        raise ValueError("If dimensions are 3 then z_col must be specified")

    # Load in data
    if dim == 2:
        df = pl.read_csv(input_file, columns=[channel_col, frame_col, x_col, y_col])
        df = df.rename(
            {channel_col: "channel", frame_col: "frame", x_col: "x", y_col: "y"}
        )
    elif dim == 3:
        df = pl.read_csv(
            input_file, columns=[channel_col, frame_col, x_col, y_col, z_col]
        )
        df = df.rename(
            {
                channel_col: "channel",
                frame_col: "frame",
                x_col: "x",
                y_col: "y",
                z_col: "z",
            }
        )

    # Specify channels to consider
    # if channel_choice is None:
    #    channels = df["channel"].unique()
    #    channels = sorted(channels)
    # else:
    channels = channel_choice

    # Get name of file - assumes last part of input file name
    name = os.path.basename(os.path.normpath(input_file)).strip(".csv")

    return datastruc.item(name, df, dim, channels)



#!/usr/bin/env python
import os
import shutil

# list of file endings to remove
file_endings = [".egg-info", "__pycache__", ".tox", ".vscode"]

# folders to recursively delete
del_folders = []

# get list of folders to delete
for root, d_names, f_names in os.walk("."):
    for dir in d_names:
        directory = root + "\\" + dir
        for file_ending in file_endings:
            if directory.endswith(file_ending):
                del_folders.append(root + "\\" + dir)

print("Folders that will be deleted!: ")
print(del_folders)

# delete folders
confirm = input("Are you sure? (yes) ")
if confirm == "yes":
    print("Deleting...")
    for folder in del_folders:
        shutil.rmtree(folder)
else:
    print("Stopped deletion.")



"""Module which allows for the watershed algorithm to
be implemented"""

# imports for widget
from PyQt5.QtWidgets import QGridLayout, QWidget, QGraphicsScene, QGraphicsView
from PyQt5.QtWidgets import QGraphicsPixmapItem, QMessageBox, QPushButton
from PyQt5.QtGui import QPixmap, QPen, QBrush
from PyQt5.QtGui import QImage, QColor
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QApplication

from skimage.segmentation import watershed
import numpy as np
import warnings


class WatershedWidget(QWidget):
    """WatershedWidget

    Widget to visualise an image as Grayscale 8 bit, then annotate with the seeds
    used for watershed.
    Child of QWidget.
    Left click places a seed, right click removes the seed.
    On closing the seed coordinates are saved

    Args:
        None

    Attributes:
        scene(QGraphicsScene): This holds all the 2D items (img, all ellipses added)
        view(QGraphicsView): Visualise the scene
        pixmap_item(QGraphicsPixmapItem): Image representation that can be used as
             a paint device
        help_button (QPushButton): Displays help for user
        pen(QPen): How the QPainter draws lines and outlines of shapes
        brush(QBrush): How to fill the shapes drawn by the painter
        marker_coords(list): List containing coordinates of the seeds for watershed

    """

    def __init__(self, img, coords=[], file_name="Image"):
        """Constructor

        Args:
             img (int8 numpy array): Numpy array range [0 255] which will be
                 converted to correct format for pixmap
             coords (list): Seed coordinate list which will pass by reference
             file_name (img name): Will display the image name
        """

        super().__init__()

        # create scene to add to and visualise it
        self.scene = QGraphicsScene()
        self.view = QGraphicsView(self.scene)
        self.setWindowTitle(file_name)

        # create layout manager add the widget to it
        self.layout = QGridLayout()
        self.layout.addWidget(self.view, 0, 0)
        self.setLayout(self.layout)  # set layout manager for widget

        # image in format pyqt needs
        # create pixmap item and add to scene
        img = img.copy()
        image = QImage(img, img.shape[1], img.shape[0], QImage.Format.Format_Grayscale8)
        self.pixmap_item = QGraphicsPixmapItem(QPixmap(image))
        self.scene.addItem(self.pixmap_item)

        # instructions button
        self.help_button = QPushButton("Help")
        self.help_button.setCheckable(True)
        self.help_button.toggle()
        self.help_button.clicked.connect(self.help_button_state)
        self.layout.addWidget(self.help_button, 0, 1)

        # overload mouse press event for pixmap item to our function
        self.pixmap_item.mousePressEvent = self.label_cell

        # set pen and brush for labels
        self.pen = QPen(QColor("red"))
        self.brush = QBrush(QColor("red"))

        # create empty set of marker coordinates
        self.marker_coords = coords

    def help_button_state(self):
        """If help button is clicked display alert widget which details the help"""

        alert = QMessageBox()
        alert.setWindowTitle("Help")
        alert.setText(
            "Left click: add a seed \n"
            "Right click: remove a seed\n"
            "When closing all seed coordinates will be saved!"
        )
        alert.exec()

    def label_cell(self, event):
        """On mouse click this definition will be called
        
        Left click: add seed
        Right click: remove seed

        Args:
            event (QEvent): Event triggered by click"""

        # add label on left click
        if event.button() == Qt.MouseButton.LeftButton:
            self.scene.addEllipse(
                event.pos().x(), event.pos().y(), 10, 10, self.pen, self.brush
            )
        # delete label on right click
        if event.button() == Qt.MouseButton.RightButton:
            # get list of items at that location
            items = self.scene.items(
                event.pos().x(),
                event.pos().y(),
                1,
                1,
                Qt.ItemSelectionMode.IntersectsItemShape,
                Qt.SortOrder.AscendingOrder,
            )
            # if item is ellipse (type 4) remove it (don't remove image!)
            for item in items:
                if item.type() == 4:
                    self.scene.removeItem(item)

    def get_coords(self):
        """On closing the widget the coordinates of the remaining seeds are extracted"""

        # perform when closed
        items = self.scene.items()
        warnings.warn(
            "Not accurate - needs to be adjusted before exact coordinate is returned"
        )

        # if ellipse add item coordinates to list
        for item in items:
            # if ellipse
            if item.type() == 4:
                # note not sure if lines up correct position - not issue here
                # i.e. if click at 100,240 is marker placed at 100,240?
                # and if its the centre of the marker? shouldn't we return
                # centre of marker
                # as the location
                # flag as issue!!!
                # but would be for accurate segmentation
                pos = item.pos()
                pos += item.rect().topLeft()
                # marker_coords is [H,W] relative to top left of image
                # i.e. if click bottom left of image marker_coords = [470,10]
                # this is what is returned to user
                self.marker_coords.append((int(pos.y()), int(pos.x())))

    def closeEvent(self, event):
        """When closing the widget this function will be overloaded.
        
        It will ask the user if they are sure, on closing the coords of the markers
        will be extracted
        
        Args:
            event (QEvent): Event triggered by click"""

        # Confirm user choice
        reply = QMessageBox()
        reply.setText("Are you sure you want to close? (Markers are saved on closing)")
        reply.setStandardButtons(
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
        )
        reply = reply.exec()
        if reply == QMessageBox.StandardButton.Yes:
            # get the coordinates of the markers
            self.get_coords()
            event.accept()
        else:
            event.ignore()


def get_markers(img, file_name="Image") -> list:
    """Get markers for the image for watershed_segment using PyQt6 widget
    Image will be converted to [0 255] greyscale image for compatibility
    with PyQt6 widget

    Args:
        img (np.ndarray): Image for which need markers for watershed
        file_name (string): Name of the file

    Returns:
        marker_coords (list): List of tuples containing coordinates of markers (h,w)
            relative to top left of image"""

    app = QApplication([])  # sys.argv if need command line inputs
    # scale img to [0 255] and convert to int8
    img = (img - np.min(img)) / (np.max(img) - np.min(img)) * 255
    img = img.astype("uint8")
    # marker coords
    marker_coords = []
    # create widget
    widget = WatershedWidget(img, coords=marker_coords, file_name=file_name)
    widget.show()
    app.exec()
    # note marker coords are (h,w) relative to top left of image i.e.
    # if click bottom left of image
    # marker_coords = [470,10]
    return marker_coords


def watershed_segment(img, file_name="Image", coords=None) -> np.ndarray:
    """Perform watershed segmentation on image - with option either to provide
    coordinates of markers (coords)
    or obtain annotation using a widget.
    
    Args:
        img (np.ndarray): Image which performing watershed on
        file_name (string): Name of the image
        coords (list): List of tuples where each tuple represents coordinate
            of a marker (x,y)
    
    Returns:
        labels (np.ndarray): Numpy array containing integer labels, each
        representing different segmented region of the input
    """

    markers = np.zeros(img.shape, dtype="int32")
    # get markers if not provided
    if coords is None:
        coords = get_markers(img, file_name)
    # coords are (h,w) in image space
    # markers[row,col] where row is h-3:h+3 and col is w-3:w+3
    # i.e. imagine if select marker in bottom left of image
    # in image space [h,w] with origin at top this would be e.g. (470,10)
    # return coordinate (470,10) from get_coords
    # markers [467:473,7:13] is populated i.e. height of 470 ish
    # and width 10 ish is populated
    # this ensures img and marker coords are in same space
    for index, coord in enumerate(coords):
        markers[coord[0] - 3 : coord[0] + 3, coord[1] - 3 : coord[1] + 3] = int(
            index + 1
        )
    # perform watershed
    labels = watershed(img, markers=markers)
    return labels



"""Various functions for visualising performance"""

import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import confusion_matrix


def plot_pr_curve(
    ax,
    label,
    linestyle,
    colour,
    precision,
    recall,
    baseline,
):
    """Plots the precision recall curve and saves

    Args:
        ax (matplotlib figure): Figure
        label (string): Label for the plot
        linestyle (string): Linestyle for the plot
        colour (string): Colour for the plot
        precision (list): List of precisions
        recall (list): List of recalls
        baseline (float): The baseline performance
            value
        pickle (bool): Whether to save the figure as pickle
            so can reload and replot

    Returns:
        None
    """

    lw = 2
    ax.plot(
        recall,
        precision,
        color=colour,
        lw=lw,
        linestyle=linestyle,
        label=label,
    )

    # baseline is just always predicting all positive class
    ax.plot([0, 1], [baseline, baseline], color="navy", lw=lw, linestyle="--")
    ax.plot([0, 1], [1, 1], color="#D41159", lw=lw, linestyle="-")
    ax.plot([1, 1], [0, 1], color="#D41159", lw=lw, linestyle="-")
    ax.set_xlabel("Recall", fontsize=18)
    ax.set_ylabel("Precision", fontsize=18)
    ax.set_xticks(np.linspace(0, 1, 11), fontsize=18)
    ax.set_yticks(np.linspace(0, 1, 11), fontsize=18)
    # this turns off the y ticks
    # plt.yticks([], [])
    ax.set_ylim((0, 1.1))
    ax.legend(loc="upper right")


def generate_conf_matrix(gtlist, predlist, classes, saveloc):
    """Generates a confusion matrix and saves it

    Args:
        gtlist (list) : List of ground truth labels
        predlist (list): List of predicted labels
        classes (list): List of classes associated
            with each label
        saveloc (string): Location to save the confusion matrix
            to

    Returns:
        None
    """
    conf_mat = confusion_matrix(gtlist, predlist)
    conf_mat = np.array(conf_mat)

    fig, ax = plt.subplots(figsize=(5, 5))
    im = ax.imshow(conf_mat, cmap="YlGn")

    # We want to show all ticks...
    ax.set_xticks(np.arange(len(classes)))
    ax.set_yticks(np.arange(len(classes)))
    # ... and label them with the respective list entries
    ax.set_xticklabels(classes)
    ax.set_yticklabels(classes)
    ax.xaxis.tick_top()
    ax.xaxis.set_label_position("top")
    plt.xlabel("Predicted class", labelpad=10)
    plt.ylabel("True class", labelpad=10)

    # Create colorbar
    cbar = ax.figure.colorbar(im, ax=ax, cmap="YlGn")
    cbar.ax.set_ylabel("", rotation=-90, va="bottom")

    # Rotate the tick labels and set their alignment.
    plt.setp(
        ax.get_yticklabels(),
        rotation=90,
        ha="center",
        va="bottom",
        rotation_mode="anchor",
    )

    # Loop over data dimensions and create text annotations
    for i in range(len(classes)):
        for j in range(len(classes)):
            print("error")
            # text = ax.text(
            #    j,
            #    i,
            #    "{:.2e}".format(conf_mat[i, j]),
            #    ha="center",
            #    va="center",
            #    color="r",
            #    size=16,
            # )

    # ax.set_title("Confusion matrix")
    fig.tight_layout()

    # Per-class accuracy
    class_accuracy = 100 * conf_mat.diagonal() / conf_mat.sum(1)
    print("Class" + " | " + "Class Accuracy")
    for i in range(len(classes)):
        print(classes[i] + " | " + str(class_accuracy[i]))

    # save figure
    plt.savefig(saveloc, dpi=600)



"""Module defining transformations to apply to data"""

from torchvision import transforms
import numpy as np


class transform:
    """Wrapper for transforms to allow input
    and label to be transformed together"""

    def __init__(self, transform_list):
        """Args:
        transform_list (list) : List of transforms to be applied"""

        self.transform = transforms.Compose(transform_list)

    def __call__(self, input, label):
        """Args:
        input (numpy array) : Input histogram
        label (numpy array) : Histogram with labels"""

        print("check")
        print(input.shape)
        print(label.shape)

        data = np.concatenate((input, label), axis=0)
        data = self.transform(data)
        input = data[:-1]
        label = data[-1:]

        return input, label



"""Module for calculating performance metrics"""

from locpix.preprocessing import datastruc
import os
import polars as pl
import numpy as np


def metric_calculation(item: datastruc.item, labels):
    """Calculate TP, TN, FP and FN for df for each label present

    Args:
        item (datastruc.item): Item containing the
            dataframe with gt and pred label
        labels (list) : List of labels to calculate results for

    Returns:
        dict : Results
            The key is the label and at each index
            is a dictionary containing TP, TN, FP, FN
            e.g. {0:{'TP': 0.6, 'FP':.3, ...}, 1:{'TP': 0.4, 'FP':.2, ...}]
            means we have two labels 0 and 1"""

    # visualise for checking
    # visualise(gt_df, pred_df, labels)

    results = {}

    # check no duplicates
    if len(item.df) != len(item.df.unique()):
        input(
            "You have duplicates in dataframe - remove this line and \
            replace with value error once appropriate"
        )

    for label in labels:
        tp_df = item.df.filter(
            (pl.col("gt_label") == label) & (pl.col("pred_label") == label)
        )
        TP = len(tp_df)
        fp_df = item.df.filter(
            (pl.col("gt_label") != label) & (pl.col("pred_label") == label)
        )
        FP = len(fp_df)
        tn_df = item.df.filter(
            (pl.col("gt_label") != label) & (pl.col("pred_label") != label)
        )
        TN = len(tn_df)
        fn_df = item.df.filter(
            (pl.col("gt_label") == label) & (pl.col("pred_label") != label)
        )
        FN = len(fn_df)

        results[label] = {"TP": TP, "FP": FP, "TN": TN, "FN": FN}

    return results


def mean_metrics(results, labels):
    """Calculate the mean metrics

    Args:
        results (dict) : Nested dictionary where each key
            is a label representing another dictionary
            containing the number of TP, TN, FP and FN
            e.g. results[2]["TP"] is number of true
            positives for label 2
        labels (list) : List of labels"""

    acc_list = []  # empty list will have length = number of labels
    iou_list = []
    recall_list = []
    for label in labels:
        TP = results[label]["TP"]
        FP = results[label]["FP"]
        TN = results[label]["TN"]
        FN = results[label]["FN"]
        acc_list.append((TP + TN) / (TP + TN + FP + FN))
        iou_list.append((TP) / (TP + FP + FN))
        recall_list.append(TP / (TP + FN))
    macc = np.mean(acc_list)
    miou = np.mean(iou_list)
    return iou_list, acc_list, recall_list, miou, macc


def aggregated_metrics(
    files_folder, save_loc, gt_label_map, add_metrics={}, metadata={}
):
    """Calculate metrics over files, where each
    file should represent a parquet file which is a
    datastruc.item

    Args:
        files_folder (string): Location in memory of the
        datastruc.item saved as a .parquet file
        save_loc (string): where to save the calculated
            metrics
        gt_label_map (dict): dictionary where each key
            is an integer and the value is the associated
            label
        metadata (dictionary): Any additional metadata
            to save"""

    # list items
    try:
        files = os.listdir(files_folder)
    except FileNotFoundError:
        raise ValueError("There should be some files to open")

    # create empty aggregated results dictionary, with each key
    # being a label
    agg_results = {}

    labels = gt_label_map.keys()
    for label in labels:
        agg_results[label] = {"TP": 0, "FP": 0, "TN": 0, "FN": 0}

    # for each file want to aggregate TP, TN, FP, FN for each class
    for file in files:

        item = datastruc.item(None, None, None, None)
        item.load_from_parquet(os.path.join(files_folder, file))

        file_results = metric_calculation(item, labels)

        # add file results to aggregated results
        for label in labels:
            agg_results[label]["TP"] += file_results[label]["TP"]
            agg_results[label]["TN"] += file_results[label]["TN"]
            agg_results[label]["FP"] += file_results[label]["FP"]
            agg_results[label]["FN"] += file_results[label]["FN"]

    # create empty results dictionary
    results = {}

    # calculate macc/miou/oacc
    iou_list, acc_list, recall_list, miou, macc = mean_metrics(agg_results, labels)
    results["iou_list"] = iou_list
    results["acc_list"] = acc_list
    results["recall_list"] = recall_list
    results["macc"] = macc
    results["miou"] = miou

    # calculate oa
    tp_running_total = 0
    fp_running_total = 0
    for label in labels:
        TP = agg_results[label]["TP"]
        FP = agg_results[label]["FP"]
        tp_running_total += TP
        fp_running_total += FP
    oacc = (tp_running_total) / (tp_running_total + fp_running_total)
    results["oacc"] = oacc

    # save to .csv in output results
    # df = pl.DataFrame(results)
    # df.write_csv(output_results)

    # sklearn to check and also speed
    pred_list = np.array([])
    gt_list = np.array([])

    for file in files:

        item = datastruc.item(None, None, None, None)
        item.load_from_parquet(os.path.join(files_folder, file))

        gt = item.df.select(pl.col("gt_label")).to_numpy()
        pred = item.df.select(pl.col("pred_label")).to_numpy()

        gt_list = np.append(gt_list, gt)
        pred_list = np.append(pred_list, pred)

    """
    print("Sklearn results")
    acc = sklearn.metrics.accuracy_score(gt_list, pred_list)
    report = sklearn.metrics.classification_report(gt_list, pred_list)
    confusion_matrix = sklearn.metrics.confusion_matrix(gt_list, pred_list)
    prec = sklearn.metrics.precision_score(gt_list, pred_list)
    recall = sklearn.metrics.recall_score(gt_list, pred_list)
    print("acc", acc)
    print("report", report)
    print("confusion matrix", confusion_matrix)
    print("prec", prec)
    print("recall", recall)
    """

    # Create text file and save overall results/configuration to this
    lines = ["Overall results", "-----------"]
    sections = [results, add_metrics, metadata]
    for section in sections:
        for key, value in section.items():
            lines.append(f"{key} : {value}")
    lines.append(f"gt label map:  {gt_label_map}")

    with open(save_loc, "w") as f:
        f.writelines("\n".join(lines))



#!/usr/bin/env python
"""Annotate module

Take in items, convert to histograms, annotate,
visualise histo mask, save the exported annotation .parquet
"""

import yaml
import os
from locpix.preprocessing import datastruc
from locpix.visualise import vis_img
import pickle as pkl
import argparse
from locpix.scripts.preprocessing import annotate_config
import tkinter as tk
from tkinter import filedialog


def main():

    parser = argparse.ArgumentParser(description="Annotate the data")
    parser.add_argument(
        "-i",
        "--project_directory",
        action="store",
        type=str,
        help="the location of the project directory",
    )
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        type=str,
        help="the location of the .yaml configuaration file\
                             for preprocessing",
    )

    args = parser.parse_args()

    # input project directory
    if args.project_directory is not None:
        project_folder = args.project_directory
    else:
        root = tk.Tk()
        root.withdraw()
        project_folder = filedialog.askdirectory(title="Project directory")

    # configuration folder
    if args.config is not None:
        # load yaml
        with open(args.config, "r") as ymlfile:
            config = yaml.safe_load(ymlfile)
            annotate_config.parse_config(config)
    else:
        config = annotate_config.config_gui()

    # list items
    input_folder = os.path.join(project_folder, "preprocess/no_gt_label")
    print(input_folder)
    try:
        files = os.listdir(input_folder)
    except FileNotFoundError:
        raise ValueError("There should be some preprocessed files to open")

    # if output directory not present create it
    output_folder = os.path.join(project_folder, "annotate/annotated")
    if os.path.exists(output_folder):
        raise ValueError(f"Cannot proceed as {output_folder} already exists")
    else:
        os.makedirs(output_folder)

    # if output directory for seg imgs not present create it
    output_seg_folder = os.path.join(project_folder, "annotate/seg_imgs")
    if os.path.exists(output_seg_folder):
        raise ValueError(f"Cannot proceed as {output_seg_folder} already exists")
    else:
        os.makedirs(output_seg_folder)

    # if output directory for seg imgs not present create it
    histo_folder = os.path.join(project_folder, "annotate/histos")
    if os.path.exists(histo_folder):
        raise ValueError(f"Cannot proceed as {histo_folder} already exists")
    else:
        os.makedirs(histo_folder)

    if config["dim"] == 2:
        histo_size = (config["x_bins"], config["y_bins"])
    elif config["dim"] == 3:
        histo_size = (config["x_bins"], config["y_bins"], config["z_bins"])
    else:
        raise ValueError("Dim should be 2 or 3")

    for file in files:
        item = datastruc.item(None, None, None, None)
        item.load_from_parquet(os.path.join(input_folder, file))

        # coord2histo
        item.coord_2_histo(histo_size, vis_interpolation=config["vis_interpolation"])

        # manual segment
        item.manual_segment()

        # save df to parquet with mapping metadata
        item.save_to_parquet(
            output_folder,
            drop_zero_label=config["drop_zero_label"],
            gt_label_map=config["gt_label_map"],
        )

        # save histogram
        save_loc = os.path.join(histo_folder, item.name + ".pkl")
        with open(save_loc, "wb") as f:
            pkl.dump(item.histo, f)

        # save images
        if config["save_img"] is True:
            save_loc = output_seg_folder
            img_dict = item.get_img_dict()
            save_loc = os.path.join(output_seg_folder, item.name + ".png")
            vis_img.visualise_seg(
                img_dict,
                item.histo_mask.T,
                item.bin_sizes,
                channels=config["vis_channels"],
                threshold=config["save_threshold"],
                how=config["save_interpolate"],
                alphas=config["alphas"],
                blend_overlays=False,
                alpha_seg=config["alpha_seg"],
                cmap_img=None,
                cmap_seg=config["cmap_seg"],
                figsize=config["fig_size"],
                origin="upper",
                save=True,
                save_loc=save_loc,
                four_colour=config["four_colour"],
                background_one_colour=config["background_one_colour"],
            )

    # save yaml file
    yaml_save_loc = os.path.join(project_folder, "annotate.yaml")
    with open(yaml_save_loc, "w") as outfile:
        yaml.dump(config, outfile)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Ilastik segmentation module

Process output of Ilastik
"""

import yaml
import os
from locpix.preprocessing import datastruc
from locpix.visualise import vis_img
import numpy as np
import pickle as pkl
import argparse
from locpix.scripts.img_seg import ilastik_output_config
import tkinter as tk
from tkinter import filedialog


def main():

    parser = argparse.ArgumentParser(description="Ilastik output")
    parser.add_argument(
        "-i",
        "--project_directory",
        action="store",
        type=str,
        help="the location of the project directory",
    )
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        type=str,
        help="the location of the .yaml configuaration file\
                             for preprocessing",
    )

    args = parser.parse_args()

    # input project directory
    if args.project_directory is not None:
        project_folder = args.project_directory
    else:
        root = tk.Tk()
        root.withdraw()
        project_folder = filedialog.askdirectory(title="Project directory")

    if args.config is not None:
        # load yaml
        with open(args.config, "r") as ymlfile:
            config = yaml.safe_load(ymlfile)
            ilastik_output_config.parse_config(config)
    elif args.configgui:
        config = ilastik_output_config.config_gui()

    # list items
    input_folder = os.path.join(project_folder, "annotate/annotated")
    try:
        files = os.listdir(input_folder)
    except FileNotFoundError:
        raise ValueError("There should be some files to open")

    # if output directory not present create it
    output_membrane_prob = os.path.join(
        project_folder, "ilastik/output/membrane/prob_map"
    )
    if os.path.exists(output_membrane_prob):
            raise ValueError(f"Cannot proceed as {output_membrane_prob} already exists")
    else:
        os.makedirs(output_membrane_prob)

    # if output directory not present create it
    output_cell_df = os.path.join(project_folder, "ilastik/output/cell/dataframe")
    if os.path.exists(output_cell_df):
            raise ValueError(f"Cannot proceed as {output_cell_df} already exists")
    else:
        os.makedirs(output_cell_df)

    # if output directory not present create it
    output_cell_img = os.path.join(project_folder, "ilastik/output/cell/img")
    if os.path.exists(output_cell_img):
            raise ValueError(f"Cannot proceed as {output_cell_img} already exists")
    else:
        os.makedirs(output_cell_img)

    for file in files:
        item = datastruc.item(None, None, None, None)
        item.load_from_parquet(os.path.join(input_folder, file))

        # load in histograms
        input_histo_folder = os.path.join(project_folder, "annotate/histos")
        histo_loc = os.path.join(input_histo_folder, item.name + ".pkl")
        with open(histo_loc, "rb") as f:
            histo = pkl.load(f)

        # ---- membrane segmentation ----

        # load in ilastik_seg
        input_membrane_prob = os.path.join(
            project_folder, "ilastik/ilastik_output/membrane/prob_map_unprocessed"
        )
        membrane_prob_mask_loc = os.path.join(input_membrane_prob, item.name + ".npy")
        ilastik_seg = np.load(membrane_prob_mask_loc)

        # ilastik_seg is [y,x,c] where channel 0 is membranes, channel 1 is inside cells
        # given we only have labels for membranes and not inside cells
        # will currently ignore
        # chanel 1
        ilastik_seg = ilastik_seg[:, :, 0]

        # save the probability map
        prob_loc = os.path.join(output_membrane_prob, item.name + ".npy")
        np.save(prob_loc, ilastik_seg)

        # ---- cell segmentation ----

        # load in ilastik_seg
        input_cell_mask = os.path.join(
            project_folder, "ilastik/ilastik_output/cell/ilastik_output_mask"
        )
        cell_mask_loc = os.path.join(input_cell_mask, item.name + ".npy")
        ilastik_seg = np.load(cell_mask_loc)

        # ilastik_seg is [y,x,c] where channel 0 is segmentation
        # where each integer represents different instance of a cell
        # i.e. 1 = one cell; 2 = different cell; etc.
        ilastik_seg = ilastik_seg[:, :, 0]

        # save instance mask to dataframe
        df = item.mask_pixel_2_coord(ilastik_seg)
        item.df = df
        item.save_to_parquet(output_cell_df, drop_zero_label=False, drop_pixel_col=True)

        # save cell segmentation image - consider only zero channel
        imgs = {key: value.T for (key, value) in histo.items()}
        save_loc = os.path.join(output_cell_img, item.name + ".png")
        vis_img.visualise_seg(
            imgs,
            ilastik_seg,
            item.bin_sizes,
            channels=[0],
            threshold=config["vis_threshold"],
            how=config["vis_interpolate"],
            blend_overlays=True,
            alpha_seg=0.5,
            origin="upper",
            save=True,
            save_loc=save_loc,
            four_colour=True,
        )

        # save yaml file
        yaml_save_loc = os.path.join(project_folder, "ilastik_output.yaml")
        with open(yaml_save_loc, "w") as outfile:
            yaml.dump(config, outfile)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
yaml2rst – A Simple Tool for Documenting YML Files
"""
#
# Copyright 2015-2019 by Hartmut Goebel <h.goebel@crazy-compilers.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import print_function
import re
import argparse
import sys
import os

__author__ = "Hartmut Goebel <h.goebel@crazy-compilers.com>"
__copyright__ = (
    "Copyright 2015-2019 by Hartmut Goebel " "<h.goebel@crazy-compilers.com>"
)
__licence__ = "GNU General Public License version 3 (GPL v3)"
__version__ = "0.3"

STATE_TEXT = 0
STATE_YAML = 1


def setup_patterns():

    # Create rst patterns, copied from `docutils/parsers/rst/states.py`
    # Function has been placed in the public domain.

    class Struct:
        """Stores data attributes for dotted-attribute access."""

        def __init__(self, **keywordargs):
            self.__dict__.update(keywordargs)

    enum = Struct()
    enum.formatinfo = {
        "parens": Struct(prefix="(", suffix=")", start=1, end=-1),
        "rparen": Struct(prefix="", suffix=")", start=0, end=-1),
        "period": Struct(prefix="", suffix=".", start=0, end=-1),
    }
    enum.formats = enum.formatinfo.keys()
    enum.sequences = [
        "arabic",
        "loweralpha",
        "upperalpha",
        "lowerroman",
        "upperroman",
    ]  # ORDERED!
    enum.sequencepats = {
        "arabic": "[0-9]+",
        "loweralpha": "[a-z]",
        "upperalpha": "[A-Z]",
        "lowerroman": "[ivxlcdm]+",
        "upperroman": "[IVXLCDM]+",
    }

    pats = {}
    pats["nonalphanum7bit"] = "[!-/:-@[-`{-~]"
    pats["alpha"] = "[a-zA-Z]"
    pats["alphanum"] = "[a-zA-Z0-9]"
    pats["alphanumplus"] = "[a-zA-Z0-9_-]"
    pats["enum"] = (
        "(%(arabic)s|%(loweralpha)s|%(upperalpha)s|%(lowerroman)s"
        "|%(upperroman)s|#)" % enum.sequencepats
    )

    for format in enum.formats:
        pats[format] = "(?P<%s>%s%s%s)" % (
            format,
            re.escape(enum.formatinfo[format].prefix),
            pats["enum"],
            re.escape(enum.formatinfo[format].suffix),
        )

    patterns = {
        "bullet": "[-+*\u2022\u2023\u2043]( +|$)",
        "enumerator": r"(%(parens)s|%(rparen)s|%(period)s)( +|$)" % pats,
    }
    for name, pat in patterns.items():
        patterns[name] = re.compile(pat)
    return patterns


PATTERNS = setup_patterns()


def get_indent(line):
    stripped_line = line.lstrip()
    indent = len(line) - len(stripped_line)
    if PATTERNS["bullet"].match(stripped_line) or PATTERNS["enumerator"].match(
        stripped_line
    ):

        indent += len(stripped_line.split(None, 1)[0]) + 1
    return indent


def get_stripped_line(line, strip_regex):
    if strip_regex:
        line = re.sub(strip_regex, "", line)
    return line


def convert(lines, strip_regex=None, yaml_strip_regex=None):
    state = STATE_TEXT
    last_text_line = ""
    last_indent = ""
    for line in lines:
        line = line.rstrip()
        if not line:
            # do not change state if the line is empty
            yield ""
        elif line.startswith("# ") or line == "#":
            if state != STATE_TEXT:
                yield ""
            line = get_stripped_line(line, strip_regex)
            line = last_text_line = line[2:]
            yield line
            last_indent = get_indent(line) * " "
            state = STATE_TEXT
        elif line == "---":
            pass
        else:
            if line.startswith("---"):
                line = line[3:]
            if state != STATE_YAML:
                if not last_text_line.endswith("::"):
                    yield last_indent + "::"
                yield ""
            line = get_stripped_line(line, yaml_strip_regex)
            yield last_indent + "  " + line
            state = STATE_YAML


def convert_text(yaml_text, strip_regex=None, yaml_strip_regex=None):
    return "\n".join(convert(yaml_text.splitlines(), strip_regex, yaml_strip_regex))


def convert_file(infilename, outfilename, strip_regex=None, yaml_strip_regex=None):
    with open(infilename) as infh:
        with open(outfilename, "w") as outfh:
            for line in convert(infh.readlines(), strip_regex, yaml_strip_regex):
                print(line.rstrip(), file=outfh)


def main(infilename, outfilename, strip_regex, yaml_strip_regex):
    if infilename == "-":
        infh = sys.stdin
    else:
        infh = open(infilename)
    if outfilename == "-":
        outfh = sys.stdout
    else:
        outfh = open(outfilename, "w")
    title = os.path.basename(infilename).removesuffix(".yaml")
    print(title, file=outfh)
    print("=" * len(title), file=outfh)
    for line in convert(infh.readlines(), strip_regex, yaml_strip_regex):
        print(line.rstrip(), file=outfh)


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "infilename", metavar="infile", help="YAML-file to read (`-` for stdin)"
    )
    parser.add_argument(
        "outfilename", metavar="outfile", help="rst-file to write (`-` for stdout)"
    )
    parser.add_argument(
        "--strip-regex",
        metavar="regex",
        help=(
            "Regex which will remove everything it matches. "
            "Can be used e.g. to remove fold markers from "
            "headings. "
            "Example to strip out [[[,]]] fold markers use: "
            r"'\s*(:?\[{3}|\]{3})\d?$'. "
            "Check the README for more details."
        ),
    )
    parser.add_argument(
        "--yaml-strip-regex",
        metavar="regex",
        help=(
            "Same usage as --strip-regex except that this "
            "regex substitution is preformed on the YAMl "
            "part of the file as opposed RST part."
        ),
    )
    args = parser.parse_args()
    main(**vars(args))


if __name__ == "__main__":
    run()



#!/usr/bin/env python
"""Cellpose segmentation module

Take in items and train the Cellpose module
"""

import yaml
import tkinter as tk
from tkinter import filedialog
import torch
from torch.utils.data import DataLoader
from locpix.img_processing.data_loading import dataset
from locpix.img_processing.training import train
import os
from torchvision import transforms
from cellpose import models
from torchsummary import summary
import argparse

# from locpix.scripts.img_seg import cellpose_train_config


def main():

    # Load in config

    parser = argparse.ArgumentParser(description="Cellpose")
    parser.add_argument(
        "-i",
        "--project_directory",
        action="store",
        type=str,
        help="the location of the project directory",
    )
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        type=str,
        help="the location of the .yaml configuaration file\
                             for preprocessing",
    )

    args = parser.parse_args()

    # input project directory
    if args.project_directory is not None:
        project_folder = args.project_directory
    else:
        root = tk.Tk()
        root.withdraw()
        project_folder = filedialog.askdirectory(title="Project directory")

    if args.config is not None:
        # load yaml
        with open(args.config, "r") as ymlfile:
            config = yaml.safe_load(ymlfile)
            # cellpose_train_config.parse_config(config)
    else:
        root = tk.Tk()
        root.withdraw()
        # gt_file_path = filedialog.askdirectory()
        # config = cellpose_train_config.config_gui(gt_file_path)

    # load in config
    input_root = os.path.join(project_folder, "annotate/annotated")
    #label_root = os.path.join(project_folder, "annotate/annotated")
    batch_size = config["batch_size"]
    epochs = config["epochs"]
    gpu = config["gpu"]
    optimiser = config["optimiser"]
    lr = config["lr"]
    weight_decay = config["weight_decay"]
    num_workers = config["num_workers"]
    loss_fn = config["loss_fn"]
    train_files = config["train_files"]

    # list items
    try:
        files = os.listdir(input_root)
        files = [os.path.splitext(file)[0] for file in files]
    except FileNotFoundError:
        raise ValueError("There should be some files to open")

    # make necessary folders if not present
    preprocessed_folder = os.path.join(project_folder, "cellpose_train")
    if os.path.exists(preprocessed_folder):
        raise ValueError(f"Cannot proceed as {preprocessed_folder} already exists")
    else:
        os.makedirs(preprocessed_folder)

    print("files", files)

    # define gpu or cpu
    # if data is on gpu then don't need to pin memory
    # and this causes errors if try
    if gpu is True and not torch.cuda.is_available():
        raise ValueError(
            "No gpu available, can run on cpu\
                         instead"
        )
    elif gpu is True and torch.cuda.is_available():
        device = torch.device("cuda")
        pin_memory = False
    elif gpu is False:
        device = torch.device("cpu")
        pin_memory = True
    else:
        raise ValueError("Specify cpu or gpu !")

    # split files into train and validation
    train_files = files[0:5]
    val_files = files[5:-1]

    # check train and test files
    print("Train files")
    print(train_files)

    # define transformations for train, test
    train_transform = [transforms.ToTensor()]
    val_transform = [transforms.ToTensor()]

    # Initialise train and val dataset
    train_set = dataset.ImgDataset(
        input_root, train_files, ".parquet", train_transform
    )
    val_set = dataset.ImgDataset(
        input_root, val_files, ".parquet", val_transform
    )

    print('Preprocessing datasets')

    # Pre-process train and val dataset
    train_set.preprocess(os.path.join(preprocessed_folder, "train"))
    val_set.preprocess(os.path.join(preprocessed_folder, "val"))

    # initialise dataloaders
    train_loader = DataLoader(
        train_set,
        batch_size=batch_size,
        shuffle=True,
        pin_memory=pin_memory,
        num_workers=num_workers,
    )

    val_loader = DataLoader(
        val_set,
        batch_size=batch_size,
        shuffle=False,
        pin_memory=pin_memory,
        num_workers=num_workers,
    )

    # intialise model
    model = models.CellposeModel(model_type=config["model"])

    # cellpose need to freeze layers/ train on intermediate output etc.

    # initialise optimiser
    if optimiser == "adam":
        optimiser = torch.optim.Adam(
            model.parameters(), lr=lr, weight_decay=weight_decay
        )

    # initialise loss function
    if loss_fn == "nll":
        loss_fn = torch.nn.functional.nll_loss

    # print parameters
    print("\n")
    print("---- Params -----")
    print("\n")
    print("Input features: ", train_set.num_node_features)
    print("Num classes: ", train_set.num_classes)
    print("Batch size: ", batch_size)
    print("Epochs: ", epochs)

    # model summary
    print("\n")
    print("---- Model summary ----")
    print("\n")
    number_nodes = 1000  # this is just for summary, has no bearing on training
    summary(
        model,
        input_size=(train_set.num_node_features, number_nodes),
        batch_size=batch_size,
    )

    # train loop
    print("\n")
    print("---- Training... ----")
    print("\n")
    train.train_loop(
        epochs, model, optimiser, train_loader, val_loader, loss_fn, device
    )
    print(
        "Need checks here to make sure model weights are\
          correct"
    )
    print("\n")
    print("---- Finished training... ----")
    print("\n")

    # save final model
    print("\n")
    print("---- Saving final model... ----")
    print("\n")


"""
    # ---------------------#

    for file in files:
        item = datastruc.item(None, None, None, None)
        item.load_from_parquet(os.path.join(config["input_folder"], file))

        # load in histograms
        histo_loc = os.path.join(config["input_histo_folder"], item.name + ".pkl")
        with open(histo_loc, "rb") as f:
            histo = pkl.load(f)

        # ---- segment membranes ----

        if config["sum_chan"] is False:
            img = histo[0].T  # consider only the zero channel
        elif config["sum_chan"] is True:
            img = histo[0].T + histo[1].T
        else:
            raise ValueError("sum_chan should be true or false")
        img = vis_img.manual_threshold(
            img, config["vis_threshold"], how=config["vis_interpolate"]
        )
        imgs = [img]
        model = models.CellposeModel(model_type=config["model"])
        channels = config["channels"]
        # note diameter is set here may want to make user choice
        # doing one at a time (rather than in batch) like this might be very slow
        _, flows, _ = model.eval(imgs, diameter=config["diameter"], channels=channels)
        # flows[0] as we have only one image so get first flow
        # flows[0][2] as this is the probability see
        # (https://cellpose.readthedocs.io/en/latest/api.html)
        semantic_mask = flows[0][2]

        # convert mask (probabilities) to range 0-1
        semantic_mask = (semantic_mask - np.min(semantic_mask)) / (
            np.max(semantic_mask) - np.min(semantic_mask)
        )

        # ---- segment cells ----
        # get markers
        markers_loc = os.path.join(config["markers_loc"], item.name + ".npy")
        try:
            markers = np.load(markers_loc)
        except FileNotFoundError:
            raise ValueError(
                "Couldn't open the file/No markers were found in relevant location"
            )

        # tested very small amount annd line below is better than
        # doing watershed on grey_log_img
        instance_mask = watershed.watershed_segment(
            semantic_mask, coords=markers
        )  # watershed on the grey image

        # ---- save ----

        # save membrane mask
        save_loc = os.path.join(config["output_membrane_prob"], item.name + ".npy")
        np.save(save_loc, semantic_mask)

        # save markers
        np.save(markers_loc, markers)

        # save instance mask to dataframe
        df = item.mask_pixel_2_coord(instance_mask)
        item.df = df
        item.save_to_parquet(
            config["output_cell_df"], drop_zero_label=False, drop_pixel_col=True
        )

        # save cell segmentation image - consider only zero channel
        imgs = {key: value.T for (key, value) in histo.items()}
        save_loc = os.path.join(config["output_cell_img"], item.name + ".png")
        vis_img.visualise_seg(
            imgs,
            instance_mask,
            item.bin_sizes,
            channels=[0],
            threshold=config["vis_threshold"],
            how=config["vis_interpolate"],
            blend_overlays=True,
            alpha_seg=0.5,
            origin="upper",
            save=True,
            save_loc=save_loc,
            four_colour=True,
        )

        # save yaml file
        yaml_save_loc = os.path.join(project_folder, "cellpose_train.yaml")
        with open(yaml_save_loc, "w") as outfile:
            yaml.dump(config, outfile)
"""

if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Ilastik prep module

Take in items and prepare for Ilastik
"""

import yaml
import os
from locpix.visualise import vis_img
import numpy as np
import pickle as pkl
import argparse
from locpix.scripts.img_seg import ilastik_prep_config
import tkinter as tk
from tkinter import filedialog


def main():

    parser = argparse.ArgumentParser(description="Ilastik prep")
    parser.add_argument(
        "-i",
        "--project_directory",
        action="store",
        type=str,
        help="the location of the project directory",
    )
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        type=str,
        help="the location of the .yaml configuaration file\
                             for preprocessing",
    )

    args = parser.parse_args()

    # input project directory
    if args.project_directory is not None:
        project_folder = args.project_directory
    else:
        root = tk.Tk()
        root.withdraw()
        project_folder = filedialog.askdirectory(title="Project directory")

    if args.config is not None:
        # load yaml
        with open(args.config, "r") as ymlfile:
            config = yaml.safe_load(ymlfile)
            ilastik_prep_config.parse_config(config)
    else:
        config = ilastik_prep_config.config_gui()

    # list items
    input_histo_folder = os.path.join(project_folder, "annotate/histos")
    try:
        files = os.listdir(input_histo_folder)
    except FileNotFoundError:
        raise ValueError("There should be some files to open")

    # if output directory not present create it
    output_folder = os.path.join(project_folder, "ilastik/prep")
    if os.path.exists(output_folder):
            raise ValueError(f"Cannot proceed as {output_folder} already exists")
    else:
        os.makedirs(output_folder)

    for file in files:

        histo_loc = os.path.join(input_histo_folder, file)

        # load in histograms
        with open(histo_loc, "rb") as f:
            histo = pkl.load(f)

        img_list = []
        for histo in histo.values():
            img = histo.T
            log_img = vis_img.manual_threshold(
                img, config["threshold"], how=config["interpolation"]
            )
            img_list.append(vis_img.img_2_grey(log_img))
        img = np.stack(img_list, axis=2)

        # all images are saved in yxc
        file_name = file.removesuffix(".pkl")
        save_loc = os.path.join(output_folder, file_name + ".npy")
        np.save(save_loc, img)

    # save yaml file
    yaml_save_loc = os.path.join(project_folder, "ilastik_prep.yaml")
    with open(yaml_save_loc, "w") as outfile:
        yaml.dump(config, outfile)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Module for evluating performance of membrane
segmentation

For the train set, it will use the probability map over
the image, and use this to assign probability of belonging
to membrane to each localisation.
Then it will use this to plot precision-recall curve,
and calculate the optimal threshold.

For the test set, it will use the probability maps to
plot a precision recall curve.
It will then use the threshold calculated from the train
set to threshold the test images

Now we have output membrane segmentations on the test set
which we can use to finally calculate the other
remaining performance metrics."""

import yaml
import os
import numpy as np
from locpix.preprocessing import datastruc
from locpix.visualise.performance import plot_pr_curve, generate_conf_matrix
import locpix.evaluate.metrics as metrics
from sklearn.metrics import precision_recall_curve, auc
import polars as pl
from datetime import datetime
import matplotlib.pyplot as plt
import pickle as pkl
from locpix.visualise import vis_img
import argparse
from locpix.scripts.img_seg import membrane_performance_config
import tkinter as tk
from tkinter import filedialog


def main():

    parser = argparse.ArgumentParser(description="Membrane performance metrics on data")
    parser.add_argument(
        "-i",
        "--project_directory",
        action="store",
        type=str,
        help="the location of the project directory",
    )
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        type=str,
        help="the location of the .yaml configuaration file\
                             for preprocessing",
    )

    args = parser.parse_args()

    # input project directory
    if args.project_directory is not None:
        project_folder = args.project_directory
    else:
        root = tk.Tk()
        root.withdraw()
        project_folder = filedialog.askdirectory(title="Project directory")

    if args.config is not None:
        # load yaml
        with open(args.config, "r") as ymlfile:
            config = yaml.safe_load(ymlfile)
            membrane_performance_config.parse_config(config)
    else:
        root = tk.Tk()
        root.withdraw()
        gt_file_path = filedialog.askdirectory(title="GT file path")
        config = membrane_performance_config.config_gui(gt_file_path)

    # list items
    gt_file_path = os.path.join(project_folder, "annotate/annotated")
    try:
        files = os.listdir(gt_file_path)
    except FileNotFoundError:
        raise ValueError("There should be some files to open")

    fig_train, ax_train = plt.subplots()
    fig_test, ax_test = plt.subplots()

    linestyles = ["dashdot", "-", "--"]
    methods = ["classic", "cellpose", "ilastik"]

    for index, method in enumerate(methods):

        print(f"{method} ...")

        # get folder names
        if method == "ilastik":
            seg_folder = os.path.join(
                project_folder, "ilastik/output/membrane/prob_map"
            )
        else:
            seg_folder = os.path.join(project_folder, f"{method}/membrane/prob_map")
        output_df_folder = os.path.join(
            project_folder, f"membrane_performance/{method}/membrane/seg_dataframes"
        )
        output_seg_imgs = os.path.join(
            project_folder, f"membrane_performance/{method}/membrane/seg_images"
        )
        output_train_pr = os.path.join(
            project_folder, f"membrane_performance/{method}/membrane/train_pr"
        )
        output_test_pr = os.path.join(
            project_folder, f"membrane_performance/{method}/membrane/test_pr"
        )
        output_metrics = os.path.join(
            project_folder, f"membrane_performance/{method}/membrane/metrics"
        )
        output_overlay_pr_curves = os.path.join(
            project_folder, "membrane_performance/overlaid_pr_curves"
        )
        output_conf_matrix = os.path.join(
            project_folder, f"membrane_performance/{method}/membrane/conf_matrix"
        )

        # create output folders
        folders = [
            output_df_folder,
            output_seg_imgs,
            output_train_pr,
            output_test_pr,
            output_metrics,
            output_overlay_pr_curves,
            output_conf_matrix,
        ]

        for folder in folders:
            # if output directory not present create it
            if os.path.exists(folder):
                raise ValueError(f"Cannot proceed as {folder} already exists")
            else:
                os.makedirs(folder)

        print("Train set...")

        prob_list = np.array([])
        gt_list = np.array([])

        for file in files:

            if file.removesuffix(".parquet") not in config["train_files"]:
                continue

            print("File ", file)

            # load df
            item = datastruc.item(None, None, None, None)
            item.load_from_parquet(os.path.join(gt_file_path, file))

            # load prob map
            img_prob = np.load(os.path.join(seg_folder, item.name + ".npy"))

            # merge prob map and df

            merged_df = item.mask_pixel_2_coord(img_prob)
            merged_df = merged_df.rename({"pred_label": "prob"})

            # get gt and predicted probability
            gt = merged_df.select(pl.col("gt_label")).to_numpy()
            prob = merged_df.select(pl.col("prob")).to_numpy()

            # append to aggregated data set
            prob_list = np.append(prob_list, prob)
            gt_list = np.append(gt_list, gt)

        # print("Sanity check... ")
        # print("gt", len(gt_list), gt_list)
        # print("pred", len(pred_list), pred_list)

        # calculate precision recall curve
        gt_list = gt_list.flatten()
        prob_list = prob_list.flatten()
        pr, rec, pr_threshold = precision_recall_curve(gt_list, prob_list, pos_label=1)
        baseline = len(gt[gt == 1]) / len(gt)

        # plot pr curve
        save_loc = os.path.join(output_train_pr, "_curve.pkl")
        plot_pr_curve(
            ax_train,
            method.capitalize(),
            linestyles[index],
            "darkorange",
            pr,
            rec,
            baseline,
            save_loc,
            pickle=True,
        )

        # calculate optimal threshold
        if config["maximise_choice"] == "recall":
            max_index = np.argmax(rec)
        elif config["maximise_choice"] == "f":
            # get f measure and find optimal
            fmeas = (2 * pr * rec) / (pr + rec)
            # remove nan from optimal threshold calculation
            nan_array = np.isnan(fmeas)
            fmeas = np.where(nan_array, 0, fmeas)
            max_index = np.argmax(fmeas)
        else:
            raise ValueError("Choice not supported")

        threshold = pr_threshold[max_index]
        print("Threshold for ", method, " :", threshold)

        print("Test set...")

        metadata = {
            "train_set": config["train_files"],
            "test_set": config["test_files"],
            "threshold": threshold,
        }

        prob_list = np.array([])
        gt_list = np.array([])
        pred_list = np.array([])

        # sanity check all have same gt label map
        gt_label_map = None

        # threshold dataframe and save to parquet file with pred label
        for file in files:

            if file.removesuffix(".parquet") not in config["test_files"]:
                continue

            print("File ", file)

            item = datastruc.item(None, None, None, None)
            item.load_from_parquet(os.path.join(gt_file_path, file))

            # load in histograms
            input_histo_folder = os.path.join(project_folder, "annotate/histos")
            histo_loc = os.path.join(input_histo_folder, item.name + ".pkl")
            with open(histo_loc, "rb") as f:
                histo = pkl.load(f)

            # load prob map
            img_prob = np.load(os.path.join(seg_folder, item.name + ".npy"))

            # merge prob map and df
            merged_df = item.mask_pixel_2_coord(img_prob)
            merged_df = merged_df.rename({"pred_label": "prob"})

            # get gt and predicted probability
            gt = merged_df.select(pl.col("gt_label")).to_numpy()
            prob = merged_df.select(pl.col("prob")).to_numpy()

            save_df = merged_df.select(
                [
                    pl.all(),
                    pl.when(pl.col("prob") > threshold)
                    .then(1)
                    .otherwise(0)
                    .alias("pred_label"),
                ]
            )

            # append to aggregated data set
            prob_list = np.append(prob_list, prob)
            gt_list = np.append(gt_list, gt)
            pred = save_df.select(pl.col("pred_label")).to_numpy()
            pred_list = np.append(pred_list, pred)

            # assign save dataframe to item
            item.df = save_df

            # save dataframe with predicted label column
            item.save_to_parquet(
                output_df_folder, drop_zero_label=False, drop_pixel_col=True
            )

            # also save image of predicted membrane
            output_img = np.where(img_prob > threshold, 1, 0)
            imgs = {key: value.T for key, value in histo.items()}

            # consider only zero channel
            save_loc = os.path.join(output_seg_imgs, item.name + ".png")
            vis_img.visualise_seg(
                imgs,
                output_img,
                item.bin_sizes,
                channels=[0],
                threshold=config["vis_threshold"],
                how=config["vis_interpolate"],
                origin="upper",
                blend_overlays=False,
                alpha_seg=0.8,
                cmap_seg=["k", "y"],
                save=True,
                save_loc=save_loc,
                four_colour=False,
            )

            # sanity check all have same gt label map
            if gt_label_map is None:
                gt_label_map = item.gt_label_map
            else:
                assert item.gt_label_map == gt_label_map

        # print("Sanity check... ")
        # print("gt", len(gt_list), gt_list)
        # print("pred", len(pred_list), pred_list)

        # calculate precision recall curve
        gt_list = gt_list.flatten()
        prob_list = prob_list.flatten()
        pr, rec, pr_threshold = precision_recall_curve(gt_list, prob_list, pos_label=1)
        baseline = len(gt[gt == 1]) / len(gt)

        # calculate confusion matrix
        date = datetime.today().strftime("%H_%M_%d_%m_%Y")
        saveloc = os.path.join(output_conf_matrix, f"conf_matrix_{date}.png")
        classes = [item.gt_label_map[0], item.gt_label_map[1]]
        pred_list = pred_list.flatten()
        generate_conf_matrix(gt_list, pred_list, classes, saveloc)
        # could just use aggregated metric function to plot the confusion matrix

        # plot pr curve
        save_loc = os.path.join(output_test_pr, "_curve.pkl")
        plot_pr_curve(
            ax_test,
            method.capitalize(),
            linestyles[index],
            "darkorange",
            pr,
            rec,
            baseline,
            save_loc,
            pickle=True,
        )
        pr_auc = auc(rec, pr)
        add_metrics = {"pr_auc": pr_auc}

        # metric calculations based on final prediction
        save_loc = os.path.join(output_metrics, f"{date}.txt")
        metrics.aggregated_metrics(
            output_df_folder,
            save_loc,
            gt_label_map,
            add_metrics=add_metrics,
            metadata=metadata,
        )

    fig_train.tight_layout()
    fig_test.tight_layout()

    # get handles and labels
    # handles_train, labels_train = ax_train.get_legend_handles_labels()
    # handles_test, labels_test = ax_test.get_legend_handles_labels()

    # specify order of items in legend
    # order = [1,0,2]

    # add legend to plot
    # ax_train.legend([handles_train[idx] for idx in order],
    #                   [methods[idx] for idx in order])
    # ax_test.legend([handles_test[idx] for idx in order],
    #                   [methods[idx] for idx in order])

    fig_train.savefig(os.path.join(output_overlay_pr_curves, "_train.png"), dpi=600)
    fig_test.savefig(os.path.join(output_overlay_pr_curves, "_test.png"), dpi=600)

    # save yaml file
    yaml_save_loc = os.path.join(project_folder, "membrane_performance.yaml")
    with open(yaml_save_loc, "w") as outfile:
        yaml.dump(config, outfile)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Cellpose segmentation module

Take in items and segment using Cellpose methods
"""

import yaml
import os
from locpix.preprocessing import datastruc
from locpix.visualise import vis_img
from locpix.img_processing import watershed
import numpy as np
import pickle as pkl
from cellpose import models
import argparse
from locpix.scripts.img_seg import cellpose_eval_config
import tkinter as tk
from tkinter import filedialog


def main():

    parser = argparse.ArgumentParser(description="Cellpose")
    parser.add_argument(
        "-i",
        "--project_directory",
        action="store",
        type=str,
        help="the location of the project directory",
    )
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        type=str,
        help="the location of the .yaml configuaration file\
                             for preprocessing",
    )
    args = parser.parse_args()

    # input project directory
    if args.project_directory is not None:
        project_folder = args.project_directory
    else:
        root = tk.Tk()
        root.withdraw()
        project_folder = filedialog.askdirectory(title="Project directory")

    # configuration
    if args.config is not None:
        # load yaml
        with open(args.config, "r") as ymlfile:
            config = yaml.safe_load(ymlfile)
            cellpose_eval_config.parse_config(config)
    else:
        config = cellpose_eval_config.config_gui()

    # list items
    input_folder = os.path.join(project_folder, "annotate/annotated")
    try:
        files = os.listdir(input_folder)
    except FileNotFoundError:
        raise ValueError("There should be some files to open")

    # output directories
    output_membrane_prob = os.path.join(project_folder, "cellpose/membrane/prob_map")
    output_cell_df = os.path.join(project_folder, "cellpose/cell/seg_dataframes")
    output_cell_img = os.path.join(project_folder, "cellpose/cell/seg_img")

    # if output directory not present create it
    output_directories = [output_membrane_prob, output_cell_df, output_cell_img]
    for directory in output_directories:
        if os.path.exists(directory):
            raise ValueError(f"Cannot proceed as {directory} already exists")
        else:
            os.makedirs(directory)

    for file in files:
        item = datastruc.item(None, None, None, None)
        input_folder = os.path.join(project_folder, "annotate/annotated")
        item.load_from_parquet(os.path.join(input_folder, file))

        # load in histograms
        input_histo_folder = os.path.join(project_folder, "annotate/histos")
        histo_loc = os.path.join(input_histo_folder, item.name + ".pkl")
        with open(histo_loc, "rb") as f:
            histo = pkl.load(f)

        # ---- segment membranes ----

        if config["sum_chan"] is False:
            img = histo[0].T  # consider only the zero channel
        elif config["sum_chan"] is True:
            img = histo[0].T + histo[1].T
        else:
            raise ValueError("sum_chan should be true or false")
        img = vis_img.manual_threshold(
            img, config["vis_threshold"], how=config["vis_interpolate"]
        )
        imgs = [img]

        model = models.CellposeModel(model_type=config["model"])
        channels = config["channels"]
        # note diameter is set here may want to make user choice
        # doing one at a time (rather than in batch) like this might be very slow
        _, flows, _ = model.eval(imgs, diameter=config["diameter"], channels=channels)
        # flows[0] as we have only one image so get first flow
        # flows[0][2] as this is the probability see
        # (https://cellpose.readthedocs.io/en/latest/api.html)
        semantic_mask = flows[0][2]

        # convert mask (probabilities) to range 0-1
        semantic_mask = (semantic_mask - np.min(semantic_mask)) / (
            np.max(semantic_mask) - np.min(semantic_mask)
        )

        # ---- segment cells ----
        # get markers
        markers_loc = os.path.join(project_folder, "markers")
        markers_loc = os.path.join(markers_loc, item.name + ".npy")
        try:
            markers = np.load(markers_loc)
        except FileNotFoundError:
            raise ValueError(
                "Couldn't open the file/No markers were found in relevant location"
            )

        # tested very small amount annd line below is better than doing
        # watershed on grey_log_img
        instance_mask = watershed.watershed_segment(
            semantic_mask, coords=markers
        )  # watershed on the grey image

        # ---- save ----

        # save membrane mask
        output_membrane_prob = os.path.join(
            project_folder, "cellpose/membrane/prob_map"
        )
        save_loc = os.path.join(output_membrane_prob, item.name + ".npy")
        np.save(save_loc, semantic_mask)

        # save markers
        np.save(markers_loc, markers)

        # save instance mask to dataframe
        df = item.mask_pixel_2_coord(instance_mask)
        item.df = df
        output_cell_df = os.path.join(project_folder, "cellpose/cell/seg_dataframes")
        item.save_to_parquet(output_cell_df, drop_zero_label=False, drop_pixel_col=True)

        # save cell segmentation image - consider only zero channel
        imgs = {key: value.T for (key, value) in histo.items()}
        output_cell_img = os.path.join(project_folder, "cellpose/cell/seg_img")
        save_loc = os.path.join(output_cell_img, item.name + ".png")
        vis_img.visualise_seg(
            imgs,
            instance_mask,
            item.bin_sizes,
            channels=[0],
            threshold=config["vis_threshold"],
            how=config["vis_interpolate"],
            blend_overlays=True,
            alpha_seg=0.5,
            origin="upper",
            save=True,
            save_loc=save_loc,
            four_colour=True,
        )

        # save yaml file
        yaml_save_loc = os.path.join(project_folder, "cellpose_eval.yaml")
        with open(yaml_save_loc, "w") as outfile:
            yaml.dump(config, outfile)


if __name__ == "__main__":
    main()



"""Preprocess config

Module which defines functiosn for managing the config
files for preprocessing.
Including:

Creating a GUI for user to choose configuration
Parsing config file to check all specified"""

from PyQt5.QtWidgets import (
    QApplication,
    QWidget,
    QMessageBox,
    QLineEdit,
    QFormLayout,
    QCheckBox,
    QListWidget,
    QPushButton,
    QFileDialog,
)
from PyQt5.QtGui import QIntValidator
from PyQt5.QtCore import Qt

import yaml

default_config_keys = [
    "x_col",
    "y_col",
    "z_col",
    "channel_col",
    "frame_col",
    "dim",
    # "output_folder",
    "channel_choice",
    "drop_pixel_col",
    "include_files",
    # "yaml_save_loc",
]


class InputWidget(QWidget):
    """Input
    Widget to take in user configuration
    Child of QWidget.
    Args:
        None
    Attributes:

    """

    def __init__(self, files, config):
        """Constructor
        Args:
            files (list) : List of files that will be preprocessed
            config (dict) : Dictionary containing the configuration
        """

        super().__init__()
        self.flo = QFormLayout()

        # Load .yaml with button
        self.load_button = QPushButton("Load yaml")
        self.load_button.clicked.connect(self.load_yaml)
        self.flo.addRow(self.load_button)

        # The following are the names of the
        # x column, y column, z column if present, channel, frame,
        # in the csvs being processed
        self.x_col = QLineEdit("X (nm)")  # x col
        self.x_col.setToolTip("Name of the x column")
        self.flo.addRow("X column", self.x_col)
        self.y_col = QLineEdit("Y (nm)")  # y col
        self.y_col.setToolTip("Name of the y column")
        self.flo.addRow("Y column", self.y_col)
        self.z_col = QLineEdit(None)  # z col
        self.z_col.setToolTip("Name of the z column (leave empty if not present)")
        self.flo.addRow("Z column", self.z_col)
        self.chan_col = QLineEdit("Channel")
        self.chan_col.setToolTip("Name of the channel column")
        self.flo.addRow("Channel column", self.chan_col)
        self.frame_col = QLineEdit("Frame")
        self.frame_col.setToolTip("Name of the frame column")
        self.flo.addRow("Frame column", self.frame_col)

        # The number of dimensions to consider
        # If 2 only deals with x and y
        # If 3 will read in and deal with z as well (currently not fully supported)
        self.dim = QLineEdit("2")
        self.dim.setToolTip(
            "The number of dimensions to consider If 2 only deals with x and y\
            If 3 will read in and deal wih\
            z as well (currently not fully supported)"
        )
        self.dim.setValidator(QIntValidator())
        self.flo.addRow("Dimensions", self.dim)

        # output folder
        # self.output_folder = QLineEdit("output/preprocessed/no_gt_label")
        # self.output_folder.setToolTip("Name of output folder")
        # self.flo.addRow("Output folder", self.output_folder)

        # choice of which channels user wants to consider
        # if null considers all
        self.channel_choice = QListWidget()
        self.channel_choice.insertItem(0, "0")
        self.channel_choice.insertItem(1, "1")
        self.channel_choice.insertItem(2, "2")
        self.channel_choice.insertItem(3, "3")
        self.channel_choice.setSelectionMode(2)
        self.channel_choice.setToolTip(
            "Choice of which channels user wants to consider"
        )
        self.flo.addRow("Channels", self.channel_choice)

        # whether to not drop the column containing
        # pixel
        self.drop_pixel_col = QCheckBox()
        self.drop_pixel_col.setToolTip(
            "whether to not drop the column containing pixel"
        )
        self.flo.addRow("Drop pixel col", self.drop_pixel_col)

        # files to include
        self.include_files = QListWidget()
        for index, value in enumerate(files):
            self.include_files.insertItem(index, value)
            self.include_files.setSelectionMode(2)
        self.include_files.setToolTip("Files to include")
        self.flo.addRow("Files to include", self.include_files)

        # yaml save loc
        # self.save_loc_input = QLineEdit("output/preprocess/preprocess.yaml")
        # self.save_loc_input.setToolTip("Yaml save location")
        # self.flo.addRow("yaml save location", self.save_loc_input)

        self.setLayout(self.flo)
        # self.include_files.selectAll()

        self.files = files
        self.config = config

    def load_yaml(self):
        """Load the yaml"""

        # Load yaml
        fname = QFileDialog.getOpenFileName(
            self, "Open file", "/home/some/folder", "Yaml (*.yaml)"
        )

        fname = str(fname[0])
        if fname != "":
            with open(fname, "r") as ymlfile:
                load_config = yaml.safe_load(ymlfile)
                if sorted(load_config.keys()) == sorted(default_config_keys):
                    self.load_config(load_config)
                else:
                    print("Can't load in as keys don't match!")

    def load_config(self, load_config):
        """Load the config into the gui

        Args:
            load_config (yaml file): Config file
                to load into the gui"""

        self.x_col.setText(load_config["x_col"])
        self.y_col.setText(load_config["y_col"])
        self.z_col.setText(load_config["z_col"])
        self.chan_col.setText(load_config["channel_col"])
        self.frame_col.setText(load_config["frame_col"])
        self.dim.setText(str(load_config["dim"]))
        # self.output_folder.setText(load_config["output_folder"])

        self.channel_choice.clearSelection()
        for chan in load_config["channel_choice"]:
            item = self.channel_choice.findItems(str(chan), Qt.MatchFlag.MatchExactly)
            item[0].setSelected(True)

        self.drop_pixel_col.setCheckState(load_config["drop_pixel_col"])

        self.include_files.clearSelection()
        for file in load_config["include_files"]:
            item = self.include_files.findItems(file, Qt.MatchFlag.MatchExactly)
            if item:
                item[0].setSelected(True)

        # self.save_loc_input.setText(load_config["yaml_save_loc"])

    def set_config(self, config):
        """Set the configuration file

        Args:
            config (dictionary) : Configuration dict"""

        config["x_col"] = self.x_col.text()
        config["y_col"] = self.y_col.text()
        config["z_col"] = self.z_col.text()
        config["channel_col"] = self.chan_col.text()
        config["frame_col"] = self.frame_col.text()
        config["dim"] = int(self.dim.text())
        # config["output_folder"] = self.output_folder.text()
        chan_list = self.channel_choice.selectedItems()
        chan_list = [int(item.text()) for item in chan_list]
        config["channel_choice"] = chan_list
        config["drop_pixel_col"] = self.drop_pixel_col.isChecked()
        include_files = self.include_files.selectedItems()
        config["include_files"] = [item.text() for item in include_files]
        # config["yaml_save_loc"] = self.save_loc_input.text()

        # check config is correct
        parse_config(config)

    def closeEvent(self, event):
        """When closing the widget this function will be overloaded.
            It will ask the user if they are sure, on closing the coords of the markers
                will be extracted
        Args:
            event (QEvent): Event triggered by click"""

        # Confirm user choice
        reply = QMessageBox()
        reply.setText("Are you sure you want to close? (Config is saved on closing)")
        reply.setStandardButtons(
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
        )
        reply = reply.exec()
        if reply == QMessageBox.StandardButton.Yes:
            self.set_config(self.config)
            event.accept()
        else:
            event.ignore()


def config_gui(files):
    """Config gui
    This function opens up a GUI for user to specify
    the configuration
    List of files files can then be unchecked if users want to ignore

    Attributes:
        files (list): List of files to be preprocessed"""

    app = QApplication([])  # sys.argv if need command line inputs
    # create widget
    config = {}
    widget = InputWidget(files, config)
    widget.show()
    app.exec()
    return config


def parse_config(config):
    """Parse config
    This function takes in the configuration .yaml
    file and checks all the necessary arguments are
    specified

    Attributes:
        config (yml file): The configuration
            .yaml file"""

    if sorted(config.keys()) != sorted(default_config_keys):
        raise ValueError(
            "Did not specify necessary default \
            configutation arguments"
        )



"""Dataset module

This module defines the dataset class for SMLM image data"""

import torch
from torch.utils.data import Dataset
import numpy as np
import os
import pickle as pkl
from locpix.preprocessing import datastruc


class ImgDataset(Dataset):
    """Pytorch dataset for the SMLM data represented as images

    Attributes:
    """

    def __init__(
        self, input_root, files, input_type, transform
    ):
        """
        Args:

            input_root (string) : Directory containing the input annotated
                SMLM data
            files (list) : List of the files to include from
                the directory in this dataset
            input_type (string) : String representing the data format of the
                input
            transform (pytorch transform) : Transforms to apply to
                the dataset
        """
        self.input_data = [
            os.path.join(input_root, file + input_type) for file in files
        ]
        #self.label_data = [
        #    os.path.join(label_root, file + label_type) for file in files
        #]
        print("input root", input_root)
        #self.input_data, self.label_data = zip(
        #    *sorted(zip(self.input_data, self.label_data))
        #)
        #print("input and label data")
        print(self.input_data)
        #print(self.label_data)
        self.transform = transform

    def preprocess(self, folder):
        """Convert the raw data into data ready for network

        Args:
            folder (string): Path containing folder to save data at
        """

        # join data
        #join_data = zip(self.input_data, self.label_data)

        # make folders to save data at if not already present
        img_folder = os.path.join(folder, 'imgs')
        label_folder = os.path.join(folder, 'labels')
        for folder in [img_folder, label_folder]:
            if os.path.exists(folder):
                raise ValueError(f"Cannot proceed as {folder} already exists")
            else:
                os.makedirs(folder)


        self.img_data = []
        self.label_data = []

        # for file in input
        for datum in self.input_data:

            # load img and label
            #with open(img, "rb") as f:
            #    histo_bad = pkl.load(f)
            
            # check img and label and check img the same
            #histos = []
            #print(type(histo_bad))
            #for key, value in histo_bad.items():
            #    histos.append(value)
            #histo_bad = np.stack(histos)
            #print("histo")
            #print(axis_2_chan)
            #print(histo.shape)
            #print(histo_bad.shape)

            #np.testing.assert_array_equal(histo, histo_bad)

            item = datastruc.item(None, None, None, None)
            item.load_from_parquet(os.path.join(datum))
            #print(item.df)

            # convert
            histo, axis_2_chan = item.render_histo()
            label = item.render_seg()

            # transpose to img space
            img = np.transpose(histo, (0,2,1))
            label = label.T

            #import matplotlib.pyplot as plt
            #plt.imshow(np.log2(img[0,:,:]), origin = 'upper', cmap = 'Greys', alpha=1)
            #plt.imshow(label, origin='upper', cmap='Reds', alpha=.4)
            #plt.show()
            #print("label", label.shape)
            #print('img', img.shape)

            # img & label path
            img_path = os.path.join(img_folder, item.name + '.npy')
            label_path = os.path.join(label_folder, item.name + '.npy')

            # add img and label path to lists
            self.img_data.append(img_path)
            self.label_data.append(label_path)

            # save img and label
            np.save(img_path, img)
            np.save(label_path, label)

    def __getitem__(self, idx):
        """Returns an item from the dataset, according to index idx

        Args:
            idx (int or other) : Index of the data to retrieve"""

        if torch.is_tensor(idx):
            idx = idx.tolist()

        input_path = self.input_data[idx]
        label_path = self.label_data[idx]

        input = np.load(input_path)
        label = np.load(label_path)

        # transform input and label together
        input, label = self.transform(input, label)

        return input, label

    def __len__(self):
        """Length of the dataset

        Args:
            None"""

        return len(self.input_data)



"""Training module.

This module contains definitions relevant to
training the model.

"""

import torch


def train_loop(epochs, model, optimiser, train_loader, val_loader, loss_fn, device):
    """Function defining the training loop

    Args:
        epochs (int): Number of epochs to train for
        optimiser (torch.optim optimiser): Optimiser
            that controls training of the model
        model (torch geometric model): Model that
            is going to be trained
        train_loader (torch dataloader): Dataloader for the
            training dataset
        val_loader (torch dataloader): Dataloader for the
            validation dataset
        loss_fn (loss function): Loss function calculate loss
            between train and output
        device (gpu or cpu): Device to train the model
            on"""

    model.to(device)

    scaler = torch.cuda.amp.GradScaler()

    for epoch in range(epochs):
        print("Epoch: ", epoch)

        # TODO : autocast look at gradient accumulation and take care with multiple gpus

        # training data
        for index, data in enumerate(train_loader):
            model.train()

            # note set to none is meant to have less memory footprint
            optimiser.zero_grad(set_to_none=True)

            # move data to device
            data.to(device)

            # forward pass - with autocasting
            with torch.autocast(device_type="cuda", dtype=torch.float64):
                # data.x = data.x.float()
                output = model(data)
                loss = loss_fn(output, data.y)

            # scales loss - calls backward on scaled loss creating scaled gradients
            scaler.scale(loss).backward()

            # unscales the gradients of optimiser then optimiser.step is called
            scaler.step(optimiser)

            # update scale for next iteration
            scaler.update()

        # val data
        # TODO: make sure torch.no_grad() somewhere
        for index, data in enumerate(val_loader):
            with torch.no_grad():

                # make sure model in eval mode
                model.eval()

                # note set to none is meant to have less memory footprint
                optimiser.zero_grad(set_to_none=True)

                # move data to device
                data.to(device)

                # forward pass - with autocasting
                with torch.autocast(device_type="cuda", dtype=torch.float16):
                    output = model(data)
                    loss = loss_fn(output, data.y)



"""Get markers config

Module which defines functions for managing the config
files for get markers.
Including:

Creating a GUI for user to choose configuration
Parsing config file to check all specified"""

from PyQt5.QtWidgets import (
    QApplication,
    QWidget,
    QMessageBox,
    QLineEdit,
    QFormLayout,
    QListWidget,
    QPushButton,
    QFileDialog,
)
from PyQt5.QtGui import QDoubleValidator
from PyQt5.QtCore import Qt
import yaml

default_config_keys = [
    "vis_threshold",
    "vis_interpolate",
]


class InputWidget(QWidget):
    """Input
    Widget to take in user configuration
    Child of QWidget.
    Args:
        None
    Attributes:

    """

    def __init__(self, config):
        """Constructor
        Args:
            files (list) : List of files that will be preprocessed
            config (dict) : Dictionary containing the configuration
        """

        super().__init__()
        self.flo = QFormLayout()

        # Load .yaml with button
        self.load_button = QPushButton("Load yaml")
        self.load_button.clicked.connect(self.load_yaml)
        self.flo.addRow(self.load_button)

        self.vis_threshold = QLineEdit("0")
        self.vis_threshold.setValidator(QDoubleValidator())
        self.flo.addRow("Vis threshold", self.vis_threshold)

        self.vis_interpolation = QListWidget()
        self.vis_interpolation.insertItem(0, "log2")
        self.vis_interpolation.insertItem(1, "log10")
        self.vis_interpolation.insertItem(2, "linear")
        self.vis_interpolation.item(0).setSelected(True)
        self.flo.addRow("vis interpolation", self.vis_interpolation)

        self.setLayout(self.flo)

        self.config = config

    def load_yaml(self):
        """Load the yaml"""

        # Load yaml
        fname = QFileDialog.getOpenFileName(
            self, "Open file", "/home/some/folder", "Yaml (*.yaml)"
        )

        fname = str(fname[0])
        if fname != "":
            with open(fname, "r") as ymlfile:
                load_config = yaml.safe_load(ymlfile)
                if sorted(load_config.keys()) == sorted(default_config_keys):
                    self.load_config(load_config)
                else:
                    print("Can't load in as keys don't match!")

    def load_config(self, load_config):
        """Load the config into the gui

        Args:
            load_config (yaml file): Config file
                to load into the gui"""

        self.vis_threshold = load_config["vis_threshold"]
        self.vis_interpolation.clearSelection()
        item = self.vis_interpolation.findItems(
            load_config["vis_interpolate"], Qt.MatchFlag.MatchExactly
        )
        item[0].setSelected(True)

    def set_config(self, config):
        """Set the configuration file

        Args:
            config (dictionary) : Configuration dict"""

        config["vis_threshold"] = float(self.vis_threshold)
        config["vis_interpolate"] = self.vis_interpolation.selectedItems()[0].text()

        # check config is correct
        parse_config(config)

    def closeEvent(self, event):
        """When closing the widget this function will be overloaded.
            It will ask the user if they are sure, on closing the coords of the markers
                will be extracted
        Args:
            event (QEvent): Event triggered by click"""

        # Confirm user choice
        reply = QMessageBox()
        reply.setText("Are you sure you want to close? (Config is saved on closing)")
        reply.setStandardButtons(
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
        )
        reply = reply.exec()
        if reply == QMessageBox.StandardButton.Yes:
            self.set_config(self.config)
            event.accept()
        else:
            event.ignore()


def config_gui():
    """Config gui
    This function opens up a GUI for user to specify
    the configuration
    List of files files can then be unchecked if users want to ignore

    Attributes:
        save_loc(string): Where to save the output
            .yaml file for the configuration
        files (list): List of files to be preprocessed"""

    app = QApplication([])  # sys.argv if need command line inputs
    # create widget
    config = {}
    widget = InputWidget(config)
    widget.show()
    app.exec()
    return config


def parse_config(config):
    """Parse config
    This function takes in the configuration .yaml
    file and checks all the necessary arguments are
    specified

    Attributes:
        config (yml file): The configuration
            .yaml file"""

    if sorted(config.keys()) != sorted(default_config_keys):
        raise ValueError(
            "Did not specify necessary default \
            configutation arguments"
        )



#!/usr/bin/env python
"""Preprocessing module

Module takes in the .csv files and processes saving the datastructures
"""

import os
import yaml
from locpix.preprocessing import functions
import argparse
import tkinter as tk
from tkinter import filedialog
from locpix.scripts.preprocessing import preprocess_config


def main():

    # load path of .csv
    parser = argparse.ArgumentParser(
        description="Preprocess the data for\
        further processing"
    )
    # data_group.add_argument(
    #    "-e",
    #    "--env",
    #    action="store",
    #    type=str,
    #    help="location of .env file for data path",
    # )
    parser.add_argument(
        "-i", "--input", action="store", type=str, help="path for the input data folder"
    )
    parser.add_argument(
       "-s",
       "--sanitycheck",
       action="store_true",
       help="whether to check correct csvs loaded in",
    )
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        type=str,
        help="the location of the .yaml configuaration file\
                             for preprocessing",
    )
    parser.add_argument(
        "-o",
        "--project_directory",
        action="store",
        type=str,
        help="the location of the project directory",
    )

    args = parser.parse_args()

    # if args.env is not None:
    #    dotenv_path = ".env"
    #    dotenv.load_dotenv(dotenv_path)
    #    csv_path = os.getenv("RAW_DATA_PATH")

    # input data folder
    if args.input is not None:
        csv_path = args.input
    else:
        root = tk.Tk()
        root.withdraw()
        csv_path = filedialog.askdirectory(title="Data folder")

    # project folder
    if args.project_directory is not None:
        project_folder = args.project_directory
    else:
        root = tk.Tk()
        root.withdraw()
        project_folder = filedialog.askdirectory(title="Project folder")

    # if output directory not present create it
    output_folder = os.path.join(project_folder, "preprocess/no_gt_label")
    if os.path.exists(output_folder):
        raise ValueError(
            "You cannot choose this project folder"
            " as it already contains preprocessed data"
        )
    else:
        os.makedirs(output_folder)

    # list all .csv at input
    csvs = os.listdir(csv_path)
    csvs = [csv for csv in csvs if csv.endswith(".csv")]

    # load in configuration .yaml
    if args.config is not None:
        # load yaml
        with open(args.config, "r") as ymlfile:
            config = yaml.safe_load(ymlfile)
            preprocess_config.parse_config(config)
    else:
        files = [csv.removesuffix(".csv") for csv in csvs]
        config = preprocess_config.config_gui(files)

    # remove excluded files
    csvs = [csv for csv in csvs if csv.removesuffix(".csv") in config["include_files"]]

    # check with user
    print("List of csvs wich will be processed")
    print(csvs)
    if args.sanitycheck:
        check = input("If you are happy with these csvs type YES: ")
        if check != "YES":
            exit()

    # go through .csv -> convert to datastructure -> save
    for csv in csvs:
        input_file = os.path.join(csv_path, csv)
        item = functions.csv_to_datastruc(
            input_file,
            config["dim"],
            config["channel_col"],
            config["frame_col"],
            config["x_col"],
            config["y_col"],
            config["z_col"],
            config["channel_choice"],
        )
        # have to not drop zero label
        # as no gt_label yet
        item.save_to_parquet(
            output_folder,
            drop_zero_label=False,
            drop_pixel_col=config["drop_pixel_col"],
        )

    # save yaml file
    config["input_data_folder"] = csv_path
    yaml_save_loc = os.path.join(project_folder, "preprocess.yaml")
    with open(yaml_save_loc, "w") as outfile:
        yaml.dump(config, outfile)


if __name__ == "__main__":
    main()



"""Module for rendering point clouds of various forms"""

import numpy as np
import polars as pl


def visualise_seg_item(item, column):
    """Function for visualising the segmentation of
    a point cloud and saving
    Where column is the column the histogram values are
    in

    Note that as not every pixel will contain localisations
    there may be pixels which get set to zero.


    Args:
        item (datastruc.item) : This contains the
            df and the labels to visualise.
            This must be true!
        column (string) : Column containing the
            segmention labels

    """

    labels = item.df.select(pl.col(column)).to_numpy()
    x_pixels = item.df.select(pl.col("x_pixel")).to_numpy()
    y_pixels = item.df.select(pl.col("y_pixel")).to_numpy()

    histo_width = item.histo[0].shape[0]
    histo_height = item.histo[0].shape[1]

    histo = np.empty((histo_width, histo_height))

    histo[x_pixels, y_pixels] = labels

    return histo
